"""Database utilities."""
from __future__ import unicode_literals, print_function, division, absolute_import
import time
import re

from sqlalchemy import Integer, event
from sqlalchemy.orm import joinedload
from sqlalchemy.schema import DDL
from sqlalchemy.sql.expression import cast, func
import transaction

from clld.db.meta import DBSession, Base
from clld.db.models import common


def as_int(col):
    return cast(col, Integer)


def with_collkey_ddl():  # pragma: no cover
    """Register creation of collkey function.

    Can be called at module level in db initialization scripts to create the collkey
    function. Once a session is bound to an engine collkey can be used to create indexes
    or in order_by clauses, e.g.::

        Index('ducet', collkey(common.Value.name)).create(DBSession.bind)
    """
    event.listen(
        Base.metadata,
        'before_create',
        DDL("""
CREATE OR REPLACE FUNCTION collkey (text, text, bool, int4, bool) RETURNS bytea
    LANGUAGE 'c' IMMUTABLE STRICT AS
    '$libdir/collkey_icu.so',
    'pgsqlext_collkey';
""").execute_if(dialect='postgresql'))


class _CollKey(object):

    """Lazily check and cache collkey support in the database on first usage."""

    _query = "SELECT EXISTS (SELECT 1 FROM pg_proc WHERE proname = 'collkey')"
    _pg_version = "SELECT current_setting('server_version_num')"

    @property
    def has_collkey(self):  # pragma: no cover
        result = DBSession.bind.dialect.name == 'postgresql'\
            and DBSession.scalar(self._query)
        self.__dict__['has_collkey'] = result  # cached overrides property
        return result

    @property
    def get_version_number(self):  # pragma: no cover
        version = DBSession.bind.dialect.name == 'postgresql'\
            and DBSession.scalar(self._pg_version)
        self.__dict__['version_number'] = int(version)
        return int(version)

    def __call__(self, col, locale='root', special_at_4=True, level=4, numeric_sorting=False):
        """If supported by the database, we use pg_collkey for collation.

        The optional arguments are passed to the collkey function as described at
        http://pgxn.org/dist/pg_collkey/0.5.1/
        """
        if self.has_collkey and (self.get_version_number < 100000):  # pragma: no cover
            return func.collkey(col, locale, special_at_4, level, numeric_sorting)

        return col


collkey = _CollKey()


def icontains(col, qs):
    """Infix search condition.

    Basic support is provided for specifying matches at beginning or end of the text.
    In addition, every special search syntax recognized by the ``LIKE`` operator of the
    underlying SQL dialect will work. Thus, to match strings with a specific <prefix>
    and <suffix>, a query string of the form "^<prefix>%<suffix>$" will do on PostgreSQL.

    .. seealso:: https://www.postgresql.org/docs/9.1/static/functions-matching.html
    """
    spattern = re.compile('^(\^|\\\\b)')
    epattern = re.compile('(\$|\\\\b)$')

    prefix, suffix = '%', '%'
    if spattern.search(qs):
        qs = spattern.sub('', qs)
        prefix = ''

    if epattern.search(qs):
        qs = epattern.sub('', qs)
        suffix = ''

    # Prevent invalid LIKE patterns:
    if qs.endswith('\\') and not qs.endswith('\\\\'):
        qs += '\\'
    return col.ilike(prefix + qs + suffix)


def compute_language_sources(*references):
    """compute relations between languages and sources.

    by going through the relevant models derived from the HasSource mixin.
    """
    old_sl = {}
    for pair in DBSession.query(common.LanguageSource):
        old_sl[(pair.source_pk, pair.language_pk)] = True

    references = list(references)
    references.extend([
        (common.ValueSetReference, 'valueset'),
        (common.SentenceReference, 'sentence')])
    sl = {}
    for model, attr in references:
        for ref in DBSession.query(model):
            sl[(ref.source_pk, getattr(ref, attr).language_pk)] = True

    for s, l in sl:
        if (s, l) not in old_sl:
            DBSession.add(common.LanguageSource(language_pk=l, source_pk=s))


def compute_number_of_values():
    """compute number of values per valueset and store it in valueset's jsondata."""
    for valueset in DBSession.query(common.ValueSet).options(
        joinedload(common.ValueSet.values)
    ):
        valueset.update_jsondata(_number_of_values=len(valueset.values))


def get_distinct_values(col, key=None):
    return sorted((c for c, in DBSession.query(col).distinct() if c), key=key)


def page_query(q, n=1000, verbose=False, commit=False):
    """Go through query results in batches.

    .. seealso:: http://stackoverflow.com/a/1217947
    """
    s = time.time()
    offset = 0
    while True:
        r = False
        for elem in q.limit(n).offset(offset):
            r = True
            yield elem
        if commit:  # pragma: no cover
            transaction.commit()
            transaction.begin()
        offset += n
        e = time.time()
        if verbose:
            print(e - s, offset, 'done')  # pragma: no cover
        s = e
        if not r:
            break


def set_alembic_version(engine, db_version):
    """Sets up the alembic_version table in an sqlite database.

    .. notes::

        This function is only to be used with sqlite databases, either when testing or
        when restoring a frozen database.

    :param engine: A connection to an sqlite database.
    :param db_version: The new version_num.
    """
    engine.execute("CREATE TABLE IF NOT EXISTS alembic_version (version_num VARCHAR(32))")
    if engine.execute("SELECT * FROM alembic_version").fetchall():  # pragma: no cover
        engine.execute("UPDATE alembic_version SET version_num = '%s'" % db_version)
    else:
        engine.execute("INSERT INTO alembic_version VALUES ('%s')" % db_version)


def get_alembic_version(engine):
    try:
        return engine.execute("SELECT version_num FROM alembic_version").fetchone()[0]
    except:  # noqa: E722; # pragma: no cover
        return None



"""
Client for the xmlrpc API of a wordpress blog.

.. note::

    we ignore blog_id altogether, see
    http://joseph.randomnetworks.com/archives/2008/06/10/\
    blog-id-in-wordpress-and-xml-rpc-blog-apis/
    thus, rely on identifying the appropriate blog by xmlrpc endpoint.
"""
import re

from six.moves import xmlrpc_client
import requests

__all__ = ['XMLRPC_PATH', 'sluggify', 'Client']
XMLRPC_PATH = 'xmlrpc.php'


def sluggify(phrase):
    phrase = phrase.lower().strip()
    phrase = re.sub('\s+', '-', phrase)
    return phrase


class Client(object):

    """Implements a client for a wpmu blog.

    provides a unified interface to functionality called over xmlrpc or plain http

    >>> c = Client('blog.example.org', 'user', 'password')
    >>> assert c.service_url == 'http://blog.example.org/xmlrpc.php'
    """

    def __init__(self, url, user, password):
        self.user = user
        self.password = password
        if not url.startswith('http://') and not url.startswith('https://'):
            url = 'http://' + url
        if not url.endswith(XMLRPC_PATH):
            if not url.endswith('/'):
                url += '/'
            url += XMLRPC_PATH
        self.service_url = url
        self.server = xmlrpc_client.Server(self.service_url)
        self.base_url = self.service_url.replace(XMLRPC_PATH, '')

    def create_post(self,
                    title,
                    content,
                    categories=None,
                    published=False,
                    date=None,
                    tags='',
                    custom_fields=None,
                    **kwargs):
        struct = dict(title=title, description=content)
        if date:
            struct['date_created_gmt'] = date
            struct['dateCreated'] = date
        if tags:
            if isinstance(tags, (list, tuple)):
                tags = ','.join(tags)
            struct['mt_keywords'] = tags
        if custom_fields is not None:
            struct['custom_fields'] = [
                dict(key=key, value=value) for key, value in custom_fields.items()]
        struct.update(kwargs)
        post_id = self.server.metaWeblog.newPost(
            '', self.user, self.password, struct, published)
        if categories:
            self.set_categories(categories, post_id)
        return post_id

    def get_categories(self, name=None):
        res = []
        for c in self.server.wp.getCategories('', self.user, self.password):
            if name:
                if c['categoryName'] == name:
                    res.append(c)
            else:
                res.append(c)
        for c in res:
            c['name'] = c['categoryName']
            c['id'] = c['categoryId']
        return res

    def set_categories(self, categories, post_id=None):
        existing_categories = dict(
            [(c['categoryName'], c) for c in self.get_categories()])
        cat_map = {}
        for cat in categories:
            if cat['name'] not in existing_categories:
                struct = dict(name=cat['name'])
                for attr in ['parent_id', 'description', 'slug']:
                    if attr in cat:
                        struct[attr] = cat[attr]
                cat_map[cat['name']] = int(
                    self.server.wp.newCategory('', self.user, self.password, struct))
            else:
                cat_map[cat['name']] = int(existing_categories[cat['name']]['id'])
        if post_id:
            self.server.mt.setPostCategories(
                post_id,
                self.user,
                self.password,
                [dict(categoryId=cat_map[name]) for name in cat_map])
        return cat_map

    def get_post_id_from_path(self, path, res=None):
        """Pretty hacky way to determine whether some post exists."""
        if path:
            if not path.startswith(self.base_url):
                path = self.base_url + path
            res = requests.get(path)
            if res.status_code != 200:
                return None
        else:
            assert res  # pragma: no cover
        m = re.search(
            """\s+name=('|")comment_post_ID('|")\s+value=('|")(?P<id>[0-9]+)('|")""",
            res.text)
        if m:
            return int(m.group('id'))
        else:
            p = '\<div\s+class\=\"post\"\s+id\=\"post\-(?P<id>[0-9]+)\"\>'
            if len(re.findall(p, res.text)) == 1:
                m = re.search(p, res.text)
                return int(m.group('id'))



"""We provide some infrastructure to build extensible database models."""
import sqlite3
import json

from six import string_types, PY2

from sqlalchemy import (
    Column, Integer, Float, String, Boolean, DateTime, func, event)
from sqlalchemy.exc import DisconnectionError
from sqlalchemy.pool import Pool
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import scoped_session, sessionmaker, deferred, undefer
from sqlalchemy.types import TypeDecorator, VARCHAR
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.orm.query import Query
from sqlalchemy.inspection import inspect
from sqlalchemy.dialects.postgresql import TSVECTOR

from zope.sqlalchemy import ZopeTransactionExtension
from clldutils.misc import NO_DEFAULT, UnicodeMixin
from clldutils import jsonlib

from clld.db.versioned import versioned_session


@event.listens_for(Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
    """Event listener to handle disconnects.

    Implements
    `pessimistic disconnect handling <http://docs.sqlalchemy.org/en/rel_0_9/core/\
    pooling.html#disconnect-handling-pessimistic>`_.

    .. note::

        Our implementation is mildly dialect specific, but works for sqlite and
        PostgreSQL. For oracle, the 'ping' query should read *SELECT 1 FROM DUAL* or
        similar.
    """
    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
        if not isinstance(dbapi_connection, sqlite3.Connection):  # pragma: no cover
            cursor.execute("SET default_text_search_config TO 'english'")
    except:  # noqa: E722; # pragma: no cover
        # dispose the whole pool instead of invalidating one at a time
        connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        raise DisconnectionError()
    cursor.close()


class ActiveOnlyQuery(Query):  # pragma: no cover

    """A pre-filtering query.

    Implements a
    `pre-filtering query <http://www.sqlalchemy.org/trac/wiki/UsageRecipes/\
    PreFilteredQuery>`_ that filters on the :py:attr:`clld.db.meta._Base.active` flag.
    """

    def get(self, ident):
        # override get() so that the flag is always checked in the
        # DB as opposed to pulling from the identity map.
        return Query.get(self.populate_existing(), ident)

    def __iter__(self):
        return Query.__iter__(self.private())

    def from_self(self, *ent):
        # override from_self() to automatically apply
        # the criterion too.   this works with count() and
        # others.
        return Query.from_self(self.private(), *ent)

    def private(self):
        mzero = self._mapper_zero()
        if mzero is not None:
            crit = mzero.class_.active == True
            return self.enable_assertions(False).filter(crit)
        else:
            return self


DBSession = scoped_session(sessionmaker(extension=ZopeTransactionExtension()))
ActiveOnlyDBSession = scoped_session(sessionmaker(
    extension=ZopeTransactionExtension(), query_cls=ActiveOnlyQuery))
VersionedDBSession = scoped_session(versioned_session(
    sessionmaker(autoflush=False, extension=ZopeTransactionExtension())))


class JSONEncodedDict(TypeDecorator):

    """Represents an immutable structure as a json-encoded string.

    Loads/serializes an empty dict for any empty value.
    """

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if not value:
            value = {}
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        if not value:
            return {}
        return json.loads(value)


class CsvMixin(object):

    """Mixin providing methods to control (de-)serialization of an object as csv row."""

    #: base name of the csv file
    __csv_name__ = None

    @classmethod
    def csv_head(cls):
        """return List of column names."""
        exclude = {'active', 'version', 'created', 'updated', 'polymorphic_type'}
        cols = sorted(
            col.key for om in inspect(cls).iterate_to_root()
            for col in om.local_table.c
            if (col.key not in exclude
                and col.type.__class__ not in [TSVECTOR]
                and not exclude.add(col.key)))
        return cols

    def value_to_csv(self, attr, ctx=None, req=None):
        """Convert one value to a representation suitable for csv writer.

        :param attr: Name of the attribute from which to convert the value.
        :return: Object suitable for serialization with csv writer.
        """
        rel = None
        if attr.endswith('__ids') or attr.endswith('__id'):
            attr = attr.split('__')
            rel = attr[-1]
            attr = '__'.join(attr[:-1])
        prop = getattr(self, attr, '')
        if attr == 'jsondata':
            prop = json.dumps(prop)
            if PY2:
                prop = prop.decode('utf8')
        if rel == 'id' and hasattr(prop, 'id'):
            return prop.id
        elif rel == 'ids':
            return ','.join('%s' % o.id for o in prop)
        return prop

    def to_csv(self, ctx=None, req=None, cols=None):
        """return list of values to be passed to csv.writer.writerow."""
        return [self.value_to_csv(attr, ctx, req) for attr in cols or self.csv_head()]

    @classmethod
    def value_from_csv(cls, attr, value):
        if not value:
            return None
        col = getattr(cls, attr)
        if hasattr(col, 'property') and hasattr(col.property, 'columns'):
            if isinstance(col.property.columns[0].type, Integer):
                return int(value)
            if isinstance(col.property.columns[0].type, Float):
                if isinstance(value, string_types):
                    value = value.replace(',', '.')
                return float(value)
        return value

    @classmethod
    def from_csv(cls, row, data=None, cols=None):
        obj = cls()
        cols = cols or obj.csv_head()
        for i, k in enumerate(cols):
            if not (k.endswith('__id') or k.endswith('__ids')) and hasattr(obj, k):
                setattr(obj, k, cls.value_from_csv(k, row[i]) or None)
        return obj

    @classmethod
    def csv_query(cls, session):
        query = session.query(cls).filter_by(active=True)
        return query.order_by(getattr(cls, 'id', getattr(cls, 'pk', None)))


class Base(UnicodeMixin, CsvMixin, declarative_base()):

    """The declarative base for all our models."""

    __abstract__ = True

    @declared_attr
    def __tablename__(cls):
        """We derive the table name from the model class name.

        This should be safe,
        because we don't want to have model classes with the same name either.
        Care has to be taken, though, to prevent collisions with the names of tables
        which are automatically created (history tables for example).
        """
        return cls.__name__.lower()

    #: All our models have an integer primary key which has nothing to do with
    #: the kind of data stored in a table. 'Natural' candidates for primary keys
    #: should be marked with unique constraints instead. This adds flexibility
    #: when it comes to database changes.
    pk = Column(Integer, primary_key=True, doc='primary key')

    #: To allow for timestamp-based versioning - as opposed or in addition to the version
    #: number approach implemented in :py:class:`clld.db.meta.Versioned` - we store
    #: a timestamp for creation or an object.
    @declared_attr
    def created(cls):
        return deferred(Column(DateTime(timezone=True), default=func.now()))

    #: Timestamp for latest update of an object.
    @declared_attr
    def updated(cls):
        return deferred(Column(DateTime(timezone=True), default=func.now(), onupdate=func.now()))

    #: The active flag is meant as an easy way to mark records as obsolete or inactive,
    #: without actually deleting them. A custom Query class could then be used which
    #: filters out inactive records.
    @declared_attr
    def active(cls):
        return deferred(Column(Boolean, default=True))

    #: To allow storage of arbitrary key,value pairs with typed values, each model
    #: provides a column to store JSON encoded dicts.
    jsondata = Column(JSONEncodedDict)

    def __init__(self, jsondata=None, **kwargs):
        kwargs['jsondata'] = jsondata or {}
        super(Base, self).__init__(**kwargs)

    def update_jsondata(self, **kw):
        """Convenience function.

        Since we use the simple
        `JSON encoded dict recipe <http://docs.sqlalchemy.org/en/rel_0_9/core/types.html\
        #marshal-json-strings>`_
        without mutation tracking, we provide a convenience method to update
        """
        d = self.jsondata.copy()
        d.update(kw)
        self.jsondata = d

    @property
    def jsondatadict(self):
        """Deprecated convenience function.

        Use jsondata directly instead, which is guaranteed to be a dictionary.
        """
        return self.jsondata or {}

    @property
    def replacement_id(self):
        """Used to allow automatically redirecting to a 'better' version of a resource."""
        if not self.active:
            return self.jsondata.get('__replacement_id__')

    @classmethod
    def get(cls, value, key=None, default=NO_DEFAULT, session=None):
        """Convenience method to query a model where exactly one result is expected.

        e.g. to retrieve an instance by primary key or id.

        :param value: The value used in the filter expression of the query.
        :param str key: The key or attribute name to be used in the filter expression. If\
        None is passed, defaults to *pk* if value is ``int`` otherwise to *id*.
        """
        session = session or DBSession
        if key is None:
            key = 'pk' if isinstance(value, int) else 'id'
        try:
            return session.query(cls)\
                .options(undefer('updated')).filter_by(**{key: value}).one()
        except (NoResultFound, MultipleResultsFound):
            if default is NO_DEFAULT:
                raise
            return default

    @classmethod
    def first(cls):
        """More convenience."""
        return DBSession.query(cls).order_by(cls.pk).first()

    def history(self):
        """return result proxy to iterate over previous versions of a record."""
        model = self.__class__
        if not hasattr(model, '__history_mapper__'):
            return []  # pragma: no cover

        history_class = model.__history_mapper__.class_
        return DBSession.query(history_class).filter(history_class.pk == self.pk)\
            .order_by(history_class.version.desc())

    def __json__(self, req):
        """Custom JSON serialization of an object.

        :param req: pyramid Request object.
        :return: ``dict`` suitable for serialization as JSON.
        """
        exclude = {'active', 'version', 'created', 'updated', 'polymorphic_type'}
        cols = [
            col.key for om in inspect(self).mapper.iterate_to_root()
            for col in om.local_table.c
            if col.key not in exclude and not exclude.add(col.key)]
        return {col: jsonlib.format(getattr(self, col)) for col in cols}

    def __unicode__(self):
        """A human readable label for the object."""
        r = getattr(self, 'name', None)
        if not r:
            r = getattr(self, 'id', None)
        if not r:
            r = '%s%s' % (self.__class__.__name__, self.pk)
        return r

    def __repr__(self):
        return '<%s %r>' % (
            self.__class__.__name__, getattr(self, 'id', self.pk))


class PolymorphicBaseMixin(object):

    """Mixin providing the wiring for joined table inheritance.

    We use joined table inheritance to allow projects to augment base ``clld``
    models with project specific attributes. This mixin class prepares
    models to serve as base classes for inheritance.
    """

    polymorphic_type = Column(String(20))

    @declared_attr
    def __mapper_args__(cls):
        return {
            'polymorphic_on': cls.polymorphic_type,
            'polymorphic_identity': 'base',
            'with_polymorphic': '*',
        }


def is_base(cls):
    """Determine whether a class is a base class or an inheriting one.

    :param cls: Model class.
    :return: ``bool`` signaling whether ``cls`` is a base class or derived, i.e.\
    customized.
    """
    # replace with inspection?
    # see http://docs.sqlalchemy.org/en/rel_0_9/orm/mapper_config.html
    # ?highlight=polymorphic_identity#sqlalchemy.orm.mapper.Mapper.polymorphic_identity
    return PolymorphicBaseMixin in cls.__bases__


class CustomModelMixin(object):

    """Mixin for customized classes in our joined table inheritance scheme.

    .. note::

        With this scheme there can be only one specialized mapper class per inheritable
        base class.
    """

    @declared_attr
    def __mapper_args__(cls):
        return {'polymorphic_identity': 'custom'}  # pragma: no cover



"""Shared functionality for clld console scripts."""
from __future__ import unicode_literals, division, absolute_import, print_function
import sys
from distutils.util import strtobool
from collections import defaultdict
import argparse
import logging
from functools import partial

from six.moves.urllib.parse import quote_plus
from six.moves import input
import transaction
from sqlalchemy import engine_from_config, create_engine
from sqlalchemy.orm import joinedload
from pyramid.paster import get_appsettings, setup_logging, bootstrap
import requests
from nameparser import HumanName
from clldutils.path import Path, as_posix, remove
from clldutils import jsonlib
from clldutils.misc import slug

from clld.db.meta import VersionedDBSession, DBSession, Base
from clld.db.models import common
from clld.db.util import page_query
from clld.lib import bibtex


def glottocodes_by_isocode(dburi, cols=['id']):
    """Query Glottolog.

    :dburi: If not None, sqlalchemy dburi for a glottolog database. If None, \
    glottolog.org will be queried.
    :cols: list of column/attribute names for which information should be gathered.
    :return: dict mapping iso639-3 codes to glottolog data.
    """
    glottocodes = {}
    if dburi:
        select = ', '.join('l.%s' % name for name in cols)
        glottolog = create_engine(dburi)
        for row in glottolog.execute(
            'select ll.hid, %s from language as l, languoid as ll where l.pk = ll.pk'
            % select
        ):
            if row[0]:
                glottocodes[row[0]] = row[1] if len(row) == 2 else row[1:]
    else:
        conv = defaultdict(lambda: lambda x: x, latitude=float, longitude=float)
        res = requests.get("http://glottolog.org/resourcemap.json?rsc=language")
        for rsc in res.json()['resources']:
            for id_ in rsc.get('identifiers', []):
                if id_['type'] == 'iso639-3':
                    row = [conv[col](rsc.get(col)) if rsc.get(col) is not None else None
                           for col in cols]
                    glottocodes[id_['identifier']] = row[0] if len(row) == 1 \
                        else tuple(row)
                    break
    return glottocodes


def add_language_codes(data, lang, isocode, glottocodes=None, glottocode=None):
    def identifier(type_, id_):
        return data.add(
            common.Identifier, '%s:%s' % (type_, id_),
            id='%s:%s' % (type_, id_),
            name=id_,
            type=getattr(common.IdentifierType, type_).value)

    if isocode and len(isocode) == 3:
        DBSession.add(common.LanguageIdentifier(
            language=lang, identifier=identifier('iso', isocode)))

    if glottocode or (glottocodes and isocode and isocode in glottocodes):
        glottocode = glottocode or glottocodes[isocode]
        DBSession.add(common.LanguageIdentifier(
            language=lang, identifier=identifier('glottolog', glottocode)))


def bibtex2source(rec, cls=common.Source, lowercase_id=False):
    year = bibtex.unescape(rec.get('year', 'nd'))
    fields = {}
    jsondata = {}
    for field in bibtex.FIELDS:
        if field in rec:
            value = bibtex.unescape(rec[field])
            container = fields if hasattr(cls, field) else jsondata
            container[field] = value

    etal = ''
    eds = ''
    authors = rec.get('author')
    if not authors:
        authors = rec.get('editor', '')
        if authors:
            eds = ' (eds.)'
    if authors:
        authors = bibtex.unescape(authors).split(' and ')
        if len(authors) > 2:
            authors = authors[:1]
            etal = ' et al.'

        authors = [HumanName(a) for a in authors]
        authors = [n.last or n.first for n in authors]
        authors = '%s%s%s' % (' and '.join(authors), etal, eds)

    return cls(
        id=slug(rec.id, lowercase=lowercase_id),
        name=('%s %s' % (authors, year)).strip(),
        description=bibtex.unescape(rec.get('title', rec.get('booktitle', ''))),
        jsondata=jsondata,
        bibtex_type=rec.genre,
        **fields)


def confirm(question, default=False):  # pragma: no cover
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    """
    while True:
        sys.stdout.write(question + (" [Y|n] " if default else " [y|N] "))
        choice = input().lower()
        if not choice:
            return default
        try:
            return strtobool(choice)
        except ValueError:
            sys.stdout.write(
                "Please respond with 'yes' or 'no' (or 'y' or 'n').\n")


def data_file(module, *comps):
    """Return Path object of file in the data directory of an app."""
    return Path(module.__file__).parent.joinpath('..', 'data', *comps)


def setup_session(config_uri, engine=None):
    setup_logging(config_uri)
    settings = get_appsettings(config_uri)
    engine = engine or engine_from_config(settings, 'sqlalchemy.')
    DBSession.configure(bind=engine)
    VersionedDBSession.configure(bind=engine)
    Base.metadata.create_all(engine)
    return Path(config_uri.split('#')[0]).resolve().parent.name


class ExistingDir(argparse.Action):  # pragma: no cover

    """Action to select an existing directory."""

    def __call__(self, parser, namespace, values, option_string=None):
        path_ = Path(values)
        if not path_.exists():
            raise argparse.ArgumentError(self, 'path does not exist')
        if not path_.is_dir():
            raise argparse.ArgumentError(self, 'path is no directory')
        setattr(namespace, self.dest, path_)


class ExistingConfig(argparse.Action):  # pragma: no cover

    """Action to select an existing config file."""

    def __call__(self, parser, namespace, values, option_string=None):
        path_ = Path(values.split('#')[0])
        if not path_.exists():
            raise argparse.ArgumentError(self, 'file does not exist')
        setattr(namespace, self.dest, values)


class SqliteDb(argparse.Action):  # pragma: no cover

    """Action to select an sqlite db to connect to."""

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, 'engine', create_engine('sqlite:///%s' % values[0]))


def parsed_args(*arg_specs, **kw):  # pragma: no cover
    """pass a truthy value as keyword parameter bootstrap to bootstrap the app."""
    parser = argparse.ArgumentParser(description=kw.pop('description', None))
    parser.add_argument(
        "config_uri", action=ExistingConfig, help="ini file providing app config")
    parser.add_argument("--glottolog-dburi", default=None)
    parser.add_argument("--module", default=None)
    parser.add_argument(
        "--sqlite", nargs=1, action=SqliteDb, help="sqlite db file")
    for args, _kw in arg_specs:
        parser.add_argument(*args, **_kw)
    args = parser.parse_args(args=kw.pop('args', None))
    engine = getattr(args, 'engine', kw.get('engine', None))
    args.env = bootstrap(args.config_uri) if kw.get('bootstrap', False) else {}
    module = setup_session(args.config_uri, engine=engine)

    # make sure we create URLs in the correct domain
    if args.env:
        dataset = DBSession.query(common.Dataset).first()
        if dataset:
            args.env['request'].environ['HTTP_HOST'] = dataset.domain

    if module == 'tests':
        module = 'clld'
    args.module = __import__(args.module or module)
    args.log = logging.getLogger(args.module.__name__)
    if engine:
        args.log.info('using bind %s' % engine)
    args.data_file = partial(data_file, args.module)
    args.module_dir = Path(args.module.__file__).parent
    args.migrations_dir = Path(args.module.__file__).parent.joinpath('..', 'migrations')
    return args


def initializedb(*args, **kw):  # pragma: no cover
    create = kw.pop('create', None)
    prime_cache = kw.pop('prime_cache', None)
    args = list(args) + [(("--prime-cache-only",), dict(action="store_true"))]
    args = parsed_args(*args, **kw)
    if not args.prime_cache_only:
        if create:
            with transaction.manager:
                create(args)
    if prime_cache:
        with transaction.manager:
            prime_cache(args)


def gbs_func(command, args, sources=None):  # pragma: no cover
    def words(s):
        return set(slug(s.strip(), remove_whitespace=False).split())

    log = args.log
    count = 0
    api_url = "https://www.googleapis.com/books/v1/volumes?"

    if command == 'cleanup':
        for fname in args.data_file('gbs').glob('*.json'):
            try:
                data = jsonlib.load(fname)
                if data.get('totalItems') == 0:
                    remove(fname)
            except ValueError:
                remove(fname)
        return

    if not sources:
        sources = DBSession.query(common.Source)\
            .order_by(common.Source.id)\
            .options(joinedload(common.Source.data))
    if callable(sources):
        sources = sources()

    for i, source in enumerate(page_query(sources, verbose=True, commit=True)):
        filepath = args.data_file('gbs', 'source%s.json' % source.id)

        if command == 'update':
            source.google_book_search_id = None
            source.update_jsondata(gbs={})

        if command in ['verify', 'update']:
            if filepath.exists():
                try:
                    data = jsonlib.load(filepath)
                except ValueError:
                    log.warn('no JSON object found in: %s' % filepath)
                    continue
                if not data['totalItems']:
                    continue
                item = data['items'][0]
            else:
                continue

        if command == 'verify':
            stitle = source.description or source.title or source.booktitle
            needs_check = False
            year = item['volumeInfo'].get('publishedDate', '').split('-')[0]
            if not year or year != slug(source.year or ''):
                needs_check = True
            twords = words(stitle)
            iwords = words(
                item['volumeInfo']['title'] + ' '
                + item['volumeInfo'].get('subtitle', ''))
            if twords == iwords \
                    or (len(iwords) > 2 and iwords.issubset(twords))\
                    or (len(twords) > 2 and twords.issubset(iwords)):
                needs_check = False
            if int(source.id) == 241:
                log.info('%s' % sorted(words(stitle)))
                log.info('%s' % sorted(iwords))
            if needs_check:
                log.info('------- %s -> %s' % (
                    source.id, item['volumeInfo'].get('industryIdentifiers')))
                log.info('%s %s' % (
                    item['volumeInfo']['title'], item['volumeInfo'].get('subtitle', '')))
                log.info(stitle)
                log.info(item['volumeInfo'].get('publishedDate'))
                log.info(source.year)
                log.info(item['volumeInfo'].get('authors'))
                log.info(source.author)
                log.info(item['volumeInfo'].get('publisher'))
                log.info(source.publisher)
                if not confirm('Are the records the same?'):
                    log.warn('---- removing ----')
                    jsonlib.dump({"totalItems": 0}, filepath)
        elif command == 'update':
            source.google_book_search_id = item['id']
            source.update_jsondata(gbs=item)
            count += 1
        elif command == 'download':
            if source.author and (source.title or source.booktitle):
                title = source.title or source.booktitle
                if filepath.exists():
                    continue
                q = [
                    'inauthor:' + quote_plus(source.author.encode('utf8')),
                    'intitle:' + quote_plus(title.encode('utf8')),
                ]
                if source.publisher:
                    q.append('inpublisher:' + quote_plus(
                        source.publisher.encode('utf8')))
                url = api_url + 'q=%s&key=%s' % ('+'.join(q), args.api_key)
                count += 1
                r = requests.get(url, headers={'accept': 'application/json'})
                log.info('%s - %s' % (r.status_code, url))
                if r.status_code == 200:
                    with open(as_posix(filepath), 'w') as fp:
                        fp.write(r.text.encode('utf8'))
                elif r.status_code == 403:
                    log.warn("limit reached")
                    break
    if command == 'update':
        log.info('assigned gbs ids for %s out of %s sources' % (count, i))
    elif command == 'download':
        log.info('queried gbs for %s sources' % count)


class Data(defaultdict):

    """Dictionary, serving to store references to new db objects during data imports.

    The values are dictionaries, keyed by the name of the model class used to create the
    new objects.

    >>> data = Data()
    >>> l = data.add(common.Language, 'l', id='abc', name='Abc Language')
    >>> assert l == data['Language']['l']
    """

    def __init__(self, **kw):
        super(Data, self).__init__(dict)
        self.defaults = kw

    def add(self, model, key, **kw):
        """
        Create an instance of a model class to be persisted in the database.

        :param model: The model class we want to create an instance of.
        :param key: A key which can be used to retrieve the instance later.
        :param kw: Keyword parameters passed to model class for initialisation.
        :return: The newly created instance of model class.
        """
        if '.' in kw.get('id', ''):
            raise ValueError('Object id contains illegal character "."')
        if list(kw.keys()) == ['_obj']:
            # if a single keyword parameter _obj is passed, we take it to be the object
            # which should be added to the session.
            new = kw['_obj']
        else:
            for k, v in self.defaults.items():
                kw.setdefault(k, v)
            new = model(**kw)
        self[model.__name__][key] = new
        DBSession.add(new)
        return new



"""Functionality to support reading and writing of excel files."""
from __future__ import unicode_literals, print_function, division, absolute_import

from six.moves import zip
import xlwt

__all__ = ['hyperlink', 'rows']


def hyperlink(url, label=None):
    f = xlwt.Font()
    f.underline = xlwt.Font.UNDERLINE_SINGLE

    style = xlwt.XFStyle()
    style.font = f
    label = label.replace('"', "'") if label else url
    return xlwt.Formula('HYPERLINK("%s";"%s")' % (url, label[:255]))


def rows(sheet, as_dict=False):
    """Read data from an excel sheet.

    :param as_dict:
        If ``True`` rows will be converted to ``dict``s using the content of the first row
        as keys.
    :return: Generator for the rows in the specified sheet.
    """
    if as_dict:
        # we use the values of the first row as keys:
        keys = [sheet.cell(0, i).value for i in range(sheet.ncols)]
        start = 1
    else:
        start = 0

    for j in range(start, sheet.nrows):
        res = [sheet.cell(j, i).value for i in range(sheet.ncols)]
        if as_dict:
            res = dict(zip(keys, res))
        yield res



"""Provides functionality to create simple SVG pie charts."""
from __future__ import unicode_literals, division, absolute_import, print_function
import math
from xml.sax.saxutils import escape

import clldutils.misc

from clld.lib.color import rgb_as_hex

__all__ = ['svg', 'data_url', 'icon', 'pie']


def svg(content, height=None, width=None):
    height = ' height="{0}"'.format(height) if height else ''
    width = ' width="{0}"'.format(width) if width else ''
    return """\
<svg  xmlns="http://www.w3.org/2000/svg"
      xmlns:xlink="http://www.w3.org/1999/xlink"{0}{1}>
  {2}
</svg>""".format(height, width, content)


def style(stroke=None, fill=None, stroke_width='1px', opacity=None):
    res = ''
    if fill:
        res += 'fill:{0};'.format(fill)
    if stroke:
        res += 'stroke:{0};stroke-width:{1};stroke-linecap:round;stroke-linejoin:round;'\
            .format(stroke, stroke_width)
    else:
        res += 'stroke:none;'
    if opacity:
        res += 'opacity:{0};'.format(opacity)
    return res


def data_url(svgxml):
    return clldutils.misc.data_url(svgxml, mimetype='image/svg+xml')


def icon(spec, opacity=None):
    paths = {
        's': 'path d="M8 8 H32 V32 H8 V8"',
        'd': 'path d="M20 2 L38 20 L20 38 L2 20 L20 2"',
        'c': 'circle cx="20" cy="20" r="14"',
        'f': 'path d="M2 4 L38 4 L20 35 L2 4"',
        't': 'path d="M2 36 L38 36 L20 5 L2 36"',
    }
    elem = '<{0} style="{1}"/>'.format(
        paths[spec[0]], style(stroke='black', fill=rgb_as_hex(spec[1:]), opacity=opacity))
    return svg(elem, height=40, width=40)


def pie(data, colors, titles=None, width=34, stroke_circle=False):
    """Create SVG pie charts.

    :return: SVG representation of the data as pie chart.
    """
    assert len(data) == len(colors)
    zipped = [(d, c) for d, c in zip(data, colors) if d != 0]
    data, colors = [z[0] for z in zipped], [z[1] for z in zipped]
    cx = cy = round(width / 2, 1)
    radius = round((width - 2) / 2, 1)
    current_angle_rad = 0
    svg_content = []
    total = sum(data)
    titles = titles or [None] * len(data)
    stroke_circle = 'black' if stroke_circle is True else stroke_circle or 'none'

    def endpoint(angle_rad):
        """
        Calculate position of point on circle given an angle, a radius, and the location
        of the center of the circle Zero line points west.
        """
        return (round(cx - (radius * math.cos(angle_rad)), 1),
                round(cy - (radius * math.sin(angle_rad)), 1))

    if len(data) == 1:
        svg_content.append(
            '<circle cx="{0}" cy="{1}" r="{2}" style="stroke:{3}; fill:{4};">'.format(
                cx, cy, radius, stroke_circle, rgb_as_hex(colors[0])))
        if titles[0]:
            svg_content.append('<title>{0}</title>'.format(escape(titles[0])))
        svg_content.append('</circle>')
        return svg(''.join(svg_content), height=width, width=width)

    for angle_deg, color, title in zip([360.0 / total * d for d in data], colors, titles):
        radius1 = "M{0},{1} L{2},{3}".format(cx, cy, *endpoint(current_angle_rad))
        current_angle_rad += math.radians(angle_deg)
        arc = "A{0},{1} 0 {2},1 {3} {4}".format(
            radius, radius, 1 if angle_deg > 180 else 0, *endpoint(current_angle_rad))
        radius2 = "L%s,%s" % (cx, cy)
        svg_content.append(
            '<path d="{0} {1} {2}" style="{3}" transform="rotate(90 {4} {5})">'.format(
                radius1, arc, radius2, style(fill=rgb_as_hex(color)), cx, cy))
        if title:
            svg_content.append('<title>{0}</title>'.format(escape(title)))
        svg_content.append('</path>')

    if stroke_circle != 'none':
        svg_content.append(
            '<circle cx="%s" cy="%s" r="%s" style="stroke:%s; fill:none;"/>'
            % (cx, cy, radius, stroke_circle))

    return svg(''.join(svg_content), height=width, width=width)



"""Generic utility functions."""
from __future__ import unicode_literals, print_function, division, absolute_import
import re
import random
from string import ascii_lowercase
from contextlib import contextmanager

from sqlalchemy.types import SchemaType, TypeDecorator, Enum
from clldutils.path import remove, move, Path
from clldutils.declenum import DeclEnum as BaseEnum
from clldutils.lgr import ABBRS as LGR_ABBRS
assert LGR_ABBRS


def random_string(length):
    return ''.join(random.choice(ascii_lowercase) for _ in range(length))


@contextmanager
def safe_overwrite(fname):
    fname = Path(fname)
    if not fname.parent.exists():
        fname.parent.mkdir()
    assert fname.parent.exists()
    tmp = fname.parent
    while tmp.exists():
        tmp = fname.parent.joinpath('%s.%s' % (fname.name, random_string(6)))
    yield tmp
    if fname.exists():
        remove(fname)
    move(tmp, fname)


def summary(text, max_length=70):
    res = ''
    words = list(reversed(text.split()))
    while words:
        nextword = words.pop()
        if len(res) + len(nextword) + 1 > max_length:
            # too long, add the word back onto the remainder
            words.append(nextword)
            break
        res += ' ' + nextword
    if words:
        res += ' ...'
    return res.strip()


# From "The Enum Recipe": http://techspot.zzzeek.org/2011/01/14/the-enum-recipe/
class DeclEnum(BaseEnum):
    @classmethod
    def db_type(cls):
        return DeclEnumType(cls)


class DeclEnumType(SchemaType, TypeDecorator):
    def __init__(self, enum):
        self.enum = enum
        self.impl = Enum(
            *enum.values(),
            name="ck%s" % re.sub(
                '([A-Z])', lambda m: "_" + m.group(1).lower(), enum.__name__))

    def _set_table(self, table, column):
        self.impl._set_table(table, column)  # pragma: no cover

    def copy(self):
        return DeclEnumType(self.enum)  # pragma: no cover

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return value.value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self.enum.from_string(value.strip())



"""
Support for the provider implementation of an OLAC OAI-PMH repository.

.. seealso:: http://www.language-archives.org/OLAC/repositories.html
"""
import re
from datetime import datetime, timedelta
from copy import copy
from collections import namedtuple

from pyramid.renderers import render
from pyramid.response import Response
from sqlalchemy.orm import joinedload_all, undefer
from clldutils.misc import UnicodeMixin

from clld.db.models.common import Language, LanguageIdentifier, Identifier, IdentifierType
from clld.interfaces import IOlacConfig


#
# OAI-PMH protocol specifics:
#
VERBS = [
    'GetRecord',
    'Identify',
    'ListIdentifiers',
    'ListMetadataFormats',
    'ListRecords',
    'ListSets',
]

ERRORS = {
    'badArgument': 'The request includes illegal arguments, is missing required '
    'arguments, includes a repeated argument, or values for arguments have an illegal '
    'syntax.',
    'badResumptionToken': 'The value of the resumptionToken argument is invalid or '
    'expired.',
    'badVerb': 'Value of the verb argument is not a legal OAI-PMH verb, the verb '
    'argument is missing, or the verb argument is repeated.',
    'cannotDisseminateFormat': 'The metadata format identified by the value given for '
    'the metadataPrefix argument is not supported by the item or by the repository.',
    'idDoesNotExist': 'The value of the identifier argument is unknown or illegal in '
    'this repository.',
    'noRecordsMatch': 'The combination of the values of the from, until, set and '
    'metadataPrefix arguments results in an empty list.',
    'noMetadataFormats': 'There are no metadata formats available for the specified '
    'item.',
    'noSetHierarchy': 'The repository does not support sets.'
}

MD_PREFIX = 'olac'

TIMESTAMP_REGEX = '[0-9]{4}\-[0-9]{2}\-[0-9]{2}'
TIMESTAMP_PATTERN = re.compile(TIMESTAMP_REGEX + '$')

Participant = namedtuple('Participant', 'role name email')
Institution = namedtuple('Institution', 'name url location')


def timestamp(dt=None):
    return str(dt or datetime.utcnow()).split('.')[0].replace(' ', 'T') + 'Z'


def date(dt=None):
    return str(dt or datetime.utcnow()).split(' ')[0]


class ResumptionToken(UnicodeMixin):

    """Represents an OAI-PMH resumption token.

    We encode all information from a List query in the resumption token so that we do
    not actually have to keep track of sequences of requests (in the spirit of REST).

    .. seealso: http://www.openarchives.org/OAI/openarchivesprotocol.html#FlowControl
    """

    PATTERN = re.compile('(?P<offset>[0-9]+)(?P<from>f%s)?(?P<until>u%s)?$'
                         % (TIMESTAMP_REGEX, TIMESTAMP_REGEX))
    limit = 100

    def __init__(self, url_arg=None, offset=None, from_=None, until=None):
        def datetime_from_iso(s):
            return datetime(*map(int, s.split('-')))

        self.offset = offset or 0
        self.from_ = from_
        self.until = until

        if url_arg is not None:
            m = self.PATTERN.match(url_arg)
            assert m
            self.offset = int(m.group('offset'))
            assert self.offset % self.limit == 0
            if m.group('from'):
                self.from_ = datetime_from_iso(m.group('from')[1:])
            if m.group('until'):
                self.until = datetime_from_iso(m.group('until')[1:]) + timedelta(1)

    def __unicode__(self):
        res = "%s" % self.offset
        if self.from_:
            res += "f%s" % date(self.from_)
        if self.until:
            res += "u%s" % date(self.until)
        assert self.PATTERN.match(res)
        return res


class OlacConfig(object):

    """Configuration of an applications OLAC repository."""

    scheme = 'oai'
    delimiter = ':'

    def _query(self, req):
        subquery = req.db.query(Identifier)\
            .filter_by(type=IdentifierType.iso.value)\
            .join(LanguageIdentifier)\
            .filter_by(language_pk=Language.pk)
        return req.db.query(Language).filter(subquery.exists())\
            .options(undefer('updated'), joinedload_all(
                Language.languageidentifier, LanguageIdentifier.identifier))

    def get_earliest_record(self, req):
        return self._query(req).order_by(Language.updated, Language.pk).first()

    def get_record(self, req, identifier):
        rec = Language.get(self.parse_identifier(req, identifier), default=None)
        assert rec
        return rec

    def query_records(self, req, from_=None, until=None):
        q = self._query(req).order_by(Language.pk)
        if from_:
            q = q.filter(Language.updated >= from_)
        if until:
            q = q.filter(Language.updated < until)
        return q

    def format_identifier(self, req, item):
        return self.delimiter.join([self.scheme, req.dataset.domain, item.id])

    def parse_identifier(self, req, id_):
        assert self.delimiter in id_
        return id_.split(self.delimiter)[-1]

    def admin(self, req):
        """Configure the archive participant with role admin.

        Note: According to http://www.language-archives.org/OLAC/repositories.html the
        list of participants
        > must include the system administrator whose email address is given in the
        > <oai:adminEmail> element of the Identify response.

        :param req: The current request.
        :return: A suitable `Participant` instance or None.
        """
        return Participant("Admin", "Archive Admin", req.dataset.contact)

    def description(self, req):
        # Note: According to http://www.language-archives.org/OLAC/repositories.html the
        # list of participants
        # > must include the system administrator whose email address is given in the
        # > <oai:adminEmail> element of the Identify response.
        participants = [self.admin(req)]
        for ed in req.dataset.editors:
            participants.append(Participant(
                "Editor",
                ed.contributor.name,
                ed.contributor.email or req.dataset.contact))
        return {
            'archiveURL': 'http://%s/' % req.dataset.domain,
            'participants': participants,
            'institution': Institution(
                req.dataset.publisher_name,
                req.dataset.publisher_url,
                req.dataset.publisher_place,
            ),
            'synopsis': req.dataset.description or '',
        }


def olac(req):
    """View implementing the OLAC OAI-PMH repository protocol."""
    return olac_with_cfg(req, req.registry.getUtility(IOlacConfig))


def olac_with_cfg(req, cfg):
    """Factory function for olac views with different configurations.

    If applications want to disseminate metadata for other resources than languages
    this function can be used to provide a second olac repository.
    """
    res = dict(verb=None, error=None, response_date=timestamp(), params={}, date=date)
    res['cfg'] = cfg

    def response(res):
        return Response(
            render('olac.mako', res, request=req).encode('utf8'),
            content_type='text/xml',
            charset=None)

    def error(_error):
        if _error:
            res['error'] = (_error, ERRORS[_error])
        return response(res)

    args = dict(req.params.items())
    res['params'] = copy(args)
    res['verb'] = args.pop('verb', None)

    if res['verb'] not in VERBS:
        return error("badVerb")

    if res['verb'] == 'ListSets':
        return error("noSetHierarchy")

    if res['verb'] == 'Identify':
        if args:
            return error("badArgument")
        res['earliest'] = res['cfg'].get_earliest_record(req)

    if res['verb'] == 'GetRecord':
        if sorted(args.keys()) != ['identifier', 'metadataPrefix']:
            return error("badArgument")

        if args['metadataPrefix'] != MD_PREFIX:
            return error("cannotDisseminateFormat")

        try:
            res['language'] = res['cfg'].get_record(req, args['identifier'])
        except AssertionError:
            return error("idDoesNotExist")

    if res['verb'] in ['ListIdentifiers', 'ListRecords']:
        if (('metadataPrefix' not in args and 'resumptionToken' not in args)
                or ('from' in args and not TIMESTAMP_PATTERN.match(args['from']))
                or ('until' in args and not TIMESTAMP_PATTERN.match(args['until']))):
            return error("badArgument")

        if [arg for arg in args if arg not in [
                'from', 'until', 'metadataPrefix', 'set', 'resumptionToken']]:
            return error("badArgument")

        if 'set' in args:
            return error("noSetHierarchy")

        if 'resumptionToken' in args:
            if len(args) > 1:
                return error("badArgument")

            try:
                rt = ResumptionToken(url_arg=args['resumptionToken'])
            except AssertionError:
                return error("badResumptionToken")
        else:
            rt = ResumptionToken(
                None, 0, args.get('from', None), args.get('until', None))

        q = res['cfg'].query_records(req, from_=rt.from_, until=rt.until)
        res['languages'] = q.offset(rt.offset).limit(rt.limit).all()
        if not res['languages']:
            return error('noRecordsMatch')

        if len(res['languages']) < rt.limit:
            res['resumptionToken'] = None
        else:
            rt.offset += rt.limit
            res['resumptionToken'] = rt

    if res['verb'] == 'ListMetadataFormats':
        if args and 'identifier' not in args:
            return error("badArgument")

    return response(res)



from __future__ import unicode_literals, print_function, division, absolute_import

from sqlalchemy import Column, String, Unicode, UniqueConstraint

from zope.interface import implementer
from nameparser import HumanName

from clld.db.meta import Base, PolymorphicBaseMixin
from clld.db.versioned import Versioned
from clld import interfaces

from .import (
    IdNameDescriptionMixin,
    DataMixin, FilesMixin, HasDataMixin, HasFilesMixin)

__all__ = ('Contributor',)


class Contributor_data(Base, Versioned, DataMixin):
    pass


class Contributor_files(Base, Versioned, FilesMixin):
    pass


@implementer(interfaces.IContributor)
class Contributor(Base,
                  PolymorphicBaseMixin,
                  Versioned,
                  IdNameDescriptionMixin,
                  HasDataMixin,
                  HasFilesMixin):

    """Creator of a contribution."""

    __table_args__ = (UniqueConstraint('name'),)

    url = Column(Unicode())
    email = Column(String)
    address = Column(Unicode)

    def last_first(self):
        if not self.name:
            return ''
        return '{0.last}, {0.first} {0.middle}'.format(HumanName(self.name)).strip()



from __future__ import unicode_literals, print_function
import json

import requests
from sqlalchemy.orm import joinedload
from six.moves.urllib.parse import quote_plus
from six import text_type
from clldutils.misc import slug

from clld.scripts.util import confirm
from clld.db.models import common
from clld.db.meta import DBSession
from clld.db.util import page_query


API_URL = 'https://archive.org/advancedsearch.php?fl[]=creator&fl[]=identifier&' \
    'fl[]=title&fl[]=year&sort[]=&sort[]=&sort[]=&rows=5&page=1&output=json&callback&' \
    'save=yes&q='


def ia_func(command, args, sources=None):  # pragma: no cover
    def words(s):
        return set(slug(s.strip(), remove_whitespace=False).split())

    log = args.log
    count = 0

    if not sources:
        sources = DBSession.query(common.Source)\
            .order_by(common.Source.id)\
            .options(joinedload(common.Source.data))
    else:
        if callable(sources):
            sources = sources()

    i = 0
    for i, source in enumerate(page_query(sources, verbose=True, commit=True)):
        filepath = args.data_file('ia', 'source%s.json' % source.id)

        if command in ['verify', 'update']:
            if filepath.exists():
                with open(filepath) as fp:
                    try:
                        data = json.load(fp)
                    except ValueError:
                        continue
                if not data['response']['numFound']:
                    continue
                item = data['response']['docs'][0]
            else:
                continue

        if command == 'verify':
            stitle = source.description or source.title or source.booktitle
            needs_check = False
            year = text_type(item.get('year', ''))
            if not year or year != slug(source.year or ''):
                needs_check = True
            twords = words(stitle)
            iwords = words(item['title'])
            if twords == iwords \
                    or (len(iwords) > 2 and iwords.issubset(twords))\
                    or (len(twords) > 2 and twords.issubset(iwords)):
                needs_check = False
            if needs_check:
                log.info('------- %s -> %s' % (source.id, item['identifier']))
                log.info(item['title'])
                log.info(stitle)
                log.info(item.get('year'))
                log.info(source.year)
                log.info(item['creator'])
                log.info(source.author)
                if not confirm('Are the records the same?'):
                    log.warn('---- removing ----')
                    with open(filepath, 'w') as fp:
                        json.dump({"response": {'numFound': 0}}, fp)
        elif command == 'update':
            source.update_jsondata(internetarchive_id=item['identifier'])
            count += 1
        elif command == 'download':
            if source.author and (source.title or source.booktitle):
                title = source.title or source.booktitle
                if filepath.exists():
                    continue
                q = quote_plus(b'creator:"%s" AND title:"%s"' % (
                    source.author.split(',')[0].encode('utf8'), title.encode('utf8')))

                count += 1
                r = requests.get(API_URL + q, headers={'accept': 'application/json'})
                log.info('%s - %s' % (r.status_code, r.url))
                if r.status_code == 200:
                    with open(filepath, 'w') as fp:
                        fp.write(r.text.encode('utf8'))
                elif r.status_code == 403:
                    log.warn("limit reached")
                    break
    if command == 'update':
        log.info('assigned internet archive identifiers for %s out of %s sources'
                 % (count, i))
    elif command == 'download':
        log.info('queried internet archive for %s sources' % count)



# update_unique_null.py - adapt composite unique constraints and nullables

from __future__ import unicode_literals

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = ''
down_revision = ''

# INSTRUCTIONS:
# - cd into the clld app directory
# - $ alembic revision -m "update unique null"
# - edit migrations/versions/<...>_update_unique_null.py
# - replace the content after the line with 'down_revision = <...>' with the content below
# - $ alembic upgrade head
# - check the output of your "update unique null" migration (will end with a DryRunException):
#   - displays which columns will be set to NOT NULL ('alter column')
#   - displays which columns will set to UNIQUE ('create constraint')
#   - displays any violating rows that will be deleted ('delete N row(s)')
# - if the changes are as intended, change 'dry=False' to 'dry=True' in <...>_update_unique_null.py
# - run alembic upgrade head once again to apply the changes to your database

UNIQUE_NULL = [
    ('contributioncontributor',
        ['contribution_pk', 'contributor_pk'], []),
    ('contributionreference',
        ['contribution_pk', 'source_pk', 'description'],
        ['description']),
    ('domainelement',
        ['parameter_pk', 'name'],
        ['name']),
    ('domainelement',
        ['parameter_pk', 'number'],
        ['number']),
    ('editor',
        ['dataset_pk', 'contributor_pk'], []),
    ('languageidentifier',
        ['language_pk', 'identifier_pk'], []),
    ('languagesource',
        ['language_pk', 'source_pk'], []),
    ('sentencereference',
        ['sentence_pk', 'source_pk', 'description'],
        ['description']),
    ('unit',
        ['language_pk', 'id'],
        ['id']),
    ('unitdomainelement',
        ['unitparameter_pk', 'name'],
        ['name']),
    ('unitdomainelement',
        ['unitparameter_pk', 'ord'],
        ['ord']),
    # NOTE: <unit, unitparameter, contribution> can have multiple values and also
    # multiple unitdomainelements
    ('unitvalue',
        [
            'unit_pk',
            'unitparameter_pk',
            'contribution_pk',
            'name',
            'unitdomainelement_pk'],
        ['contribution_pk', 'name', 'unitdomainelement_pk']),
    # NOTE: <language, parameter, contribution> can have multiple values and also
    # multiple domainelements
    ('value',
        ['valueset_pk', 'name', 'domainelement_pk'],
        ['name', 'domainelement_pk']),
    ('valuesentence',
        ['value_pk', 'sentence_pk'], []),
    ('valueset',
        ['language_pk', 'parameter_pk', 'contribution_pk'],
        ['contribution_pk']),
    ('valuesetreference',
        ['valueset_pk', 'source_pk', 'description'],
        ['description']),
]


class DryRunException(Exception):
    """Raised at the end of a dry run so the database transaction is not comitted."""


def upgrade(dry=True, verbose=True):
    conn = op.get_bind()

    assert conn.dialect.name == 'postgresql'

    def delete_null_duplicates(tablename, columns, notnull, returning=sa.text('*')):
        assert columns
        table = sa.table(tablename, *map(sa.column, ['pk'] + columns))
        any_null = sa.or_(table.c[n] == sa.null() for n in notnull)
        yield table.delete(bind=conn).where(any_null).returning(returning)
        other = table.alias()
        yield table.delete(bind=conn).where(~any_null).returning(returning)\
            .where(
                sa.exists()
                .where(sa.and_(table.c[c] == other.c[c] for c in columns))
                .where(table.c.pk > other.c.pk))

    def print_rows(rows, verbose=verbose):
        if not verbose:
            return
        for r in rows:
            print('    %r' % dict(r))

    class regclass(sa.types.UserDefinedType):
        def get_col_spec(self):
            return 'regclass'

    pga = sa.table(
        'pg_attribute', *map(sa.column, ['attrelid', 'attname', 'attnum', 'attnotnull']))

    select_nullable = sa.select([pga.c.attname], bind=conn)\
        .where(pga.c.attrelid == sa.cast(sa.bindparam('table'), regclass))\
        .where(pga.c.attname == sa.func.any(sa.bindparam('notnull')))\
        .where(~pga.c.attnotnull)\
        .order_by(pga.c.attnum)

    pgco = sa.table('pg_constraint', *map(sa.column,
                    ['oid', 'conname', 'contype', 'conrelid', 'conkey']))

    sq = sa.select([
        pgco.c.conname.label('name'),
        sa.func.pg_get_constraintdef(pgco.c.oid).label('definition'),
        sa.func.array(
            sa.select([sa.cast(pga.c.attname, sa.Text)])
            .where(pga.c.attrelid == pgco.c.conrelid)
            .where(pga.c.attnum == sa.func.any(pgco.c.conkey))
            .as_scalar()).label('names'),
    ])\
        .where(pgco.c.contype == 'u')\
        .where(pgco.c.conrelid == sa.cast(sa.bindparam('table'), regclass))\
        .alias()

    select_const = sa.select([sq.c.name, sq.c.definition], bind=conn)\
        .where(sq.c.names.op('@>')(sa.bindparam('cols')))\
        .where(sq.c.names.op('<@')(sa.bindparam('cols')))

    for table, unique, null in UNIQUE_NULL:
        print(table)
        notnull = [u for u in unique if u not in null]
        delete_null, delete_duplicates = delete_null_duplicates(table, unique, notnull)

        nulls = delete_null.execute().fetchall()
        if nulls:
            print('%s delete %d row(s) violating NOT NULL(%s)' % (
                table, len(nulls), ', '.join(notnull)))
            print_rows(nulls)

        duplicates = delete_duplicates.execute().fetchall()
        if duplicates:
            print('%s delete %d row(s) violating UNIQUE(%s)' % (
                table, len(duplicates), ', '.join(unique)))
            print_rows(duplicates)

        for col, in select_nullable.execute(table=table, notnull=notnull):
            print('%s alter column %s NOT NULL' % (table, col))
            op.alter_column(table, col, nullable=False)

        constraint = 'UNIQUE (%s)' % ', '.join(unique)
        matching = select_const.execute(table=table, cols=unique).fetchall()
        if matching:
            assert len(matching) == 1
            (name, definition), = matching
            if definition == constraint:
                print('%s keep constraint %s %s\n' % (table, name, definition))
                continue
            print('%s drop constraint %s %s' % (table, name, definition))
            op.drop_constraint(name, table)
        name = '%s_%s_key' % (table, '_'.join(unique))
        print('%s create constraint %s %s' % (table, name, constraint))
        op.create_unique_constraint(name, table, unique)
        print('')

    if dry:
        raise DryRunException('set dry=False to apply these changes')


def downgrade():
    pass



"""Adapters to render bibliographic information in various formats."""
from clld.web.adapters.base import Representation
from clld.lib.bibtex import IDatabase, IRecord, Database
from clld.interfaces import IIndex, ISource, IDataTable


class _Format(Representation):

    """Virtual base class."""

    def filename(self, ctx, req):
        return '%s-refs.%s' % (req.dataset.id, self.extension)

    def render_to_response(self, ctx, req):
        res = super(_Format, self).render_to_response(ctx, req)
        res.content_disposition = 'attachment; filename="%s"' % self.filename(ctx, req)
        return res

    def render(self, ctx, req):
        if ISource.providedBy(ctx):
            ctx = ctx.bibtex()
        elif IDataTable.providedBy(ctx):
            ctx = Database([s.bibtex() for s in ctx.get_query()])
        return ctx.format(self.extension)


class Bibtex(_Format):

    """BibTeX.

    .. seealso:: http://en.wikipedia.org/wiki/BibTeX
    """

    extension = 'bib'
    mimetype = 'text/x-bibtex'


class Endnote(_Format):

    """Endnote.

    .. seealso:: http://en.wikipedia.org/wiki/EndNote
    """

    extension = 'en'
    mimetype = "application/x-endnote-refer"


class ReferenceManager(_Format):

    """ReferenceManager.

    .. seealso:: http://en.wikipedia.org/wiki/RIS_%28file_format%29
    """

    extension = 'ris'
    mimetype = "application/x-research-info-systems"


class Mods(_Format):

    """Metadata Object Description Schema.

    .. seealso:: http://www.loc.gov/standards/mods/
    """

    extension = 'mods'
    mimetype = 'application/mods+xml'

    def filename(self, ctx, req):
        return '%s-refs.mods.xml' % req.dataset.id


def includeme(config):
    for adapter in [Bibtex, Endnote, ReferenceManager, Mods]:
        for interface in [IDatabase, IRecord, ISource]:
            config.register_adapter(adapter, interface, name=adapter.extension)
        config.register_adapter(
            adapter, ISource, IIndex, name=adapter.extension)



"""Adapters to serialize metadata of clld objects."""
from itertools import chain

from zope.interface import implementer

from clld import interfaces
from clld.web.adapters.base import Representation
from clld.lib import bibtex


class Metadata(Representation):

    """Virtual base class."""

    rel = 'describedby'

    @property
    def unapi_name(self):
        return getattr(self, 'unapi', self.extension)


class MetadataFromRec(Metadata):

    """Virtual base class deriving metadata from a bibtex record."""

    def rec(self, ctx, req):
        data = {}
        if interfaces.IContribution.providedBy(ctx):
            genre = 'incollection'
            data['author'] = [
                c.name for c in
                chain(ctx.primary_contributors, ctx.secondary_contributors)]
            data['booktitle'] = req.dataset.description
            data['editor'] = [c.contributor.name for c in req.dataset.editors]
            id_ = '%s-%s' % (req.dataset.id, ctx.id)
        else:
            genre = 'book'
            data['editor'] = [c.contributor.name for c in ctx.editors]
            id_ = req.dataset.id

        return bibtex.Record(
            genre,
            id_,
            title=getattr(ctx, 'citation_name', ctx.__unicode__()),
            url=req.resource_url(ctx),
            address=req.dataset.publisher_place,
            publisher=req.dataset.publisher_name,
            year=str(req.dataset.published.year),
            **data)


@implementer(interfaces.IRepresentation, interfaces.IMetadata)
class BibTex(MetadataFromRec):

    """Resource metadata as BibTex record."""

    name = 'BibTeX'
    __label__ = 'BibTeX'
    unapi = 'bibtex'
    extension = 'md.bib'
    mimetype = 'text/x-bibtex'

    def render(self, ctx, req):
        return self.rec(ctx, req).__unicode__()


@implementer(interfaces.IRepresentation, interfaces.IMetadata)
class ReferenceManager(MetadataFromRec):

    """Resource metadata in RIS format."""

    name = 'RIS'
    __label__ = 'RIS'
    unapi = 'ris'
    extension = 'md.ris'
    mimetype = "application/x-research-info-systems"

    def render(self, ctx, req):
        return self.rec(ctx, req).format('ris')


@implementer(interfaces.IRepresentation, interfaces.IMetadata)
class TxtCitation(Metadata):

    """Resource metadata formatted as plain text citation."""

    name = "Citation"
    __label__ = 'Text'
    extension = 'md.txt'
    mimetype = 'text/plain'

    def render(self, ctx, req):
        if interfaces.IContribution.providedBy(ctx):
            self.template = 'contribution/md_txt.mako'
        else:  # if interfaces.IDataset.providedBy(ctx):
            self.template = 'dataset/md_txt.mako'
        return super(TxtCitation, self).render(ctx, req)



"""Represent clld objects as excel spreadsheets."""
from six import BytesIO
import xlwt

from clld.web.adapters.base import Index
from clld.lib.excel import hyperlink

QUERY_LIMIT = 2000


class ExcelAdapter(Index):

    """Represent tables as excel sheets (maximal %d rows)."""

    __doc__ %= QUERY_LIMIT

    extension = 'xls'
    mimetype = 'application/vnd.ms-excel'

    def header(self, ctx, req):
        return ['ID', 'Name']

    def row(self, ctx, req, item):
        return [item.id, hyperlink(req.resource_url(item), item.__unicode__())]

    def render(self, ctx, req):
        if not xlwt:
            return ''  # pragma: no cover
        wb = xlwt.Workbook()
        ws = wb.add_sheet(ctx.__unicode__())

        for i, col in enumerate(self.header(ctx, req)):
            ws.write(0, i, col)

        for j, item in enumerate(ctx.get_query(limit=QUERY_LIMIT)):
            for i, col in enumerate(self.row(ctx, req, item)):
                ws.write(j + 1, i, col)

        out = BytesIO()
        wb.save(out)
        out.seek(0)
        return out.read()

    def render_to_response(self, ctx, req):
        res = super(ExcelAdapter, self).render_to_response(ctx, req)
        res.content_disposition = 'attachment; filename="%s.xls"' % repr(ctx)
        return res


class Languages(ExcelAdapter):

    """Represent DataTable of Language instances as excel sheet (maximal %d rows)."""

    __doc__ %= QUERY_LIMIT

    def header(self, ctx, req):
        return super(Languages, self).header(ctx, req) + ['Latitude', 'Longitude']

    def row(self, ctx, req, item):
        res = super(Languages, self).row(ctx, req, item)
        res.extend([item.latitude, item.longitude])
        return res


class Values(ExcelAdapter):

    """Represent table of Value instances as excel sheet (maximal %d rows)."""

    __doc__ %= QUERY_LIMIT

    def header(self, ctx, req):
        return super(Values, self).header(ctx, req) + [
            'Parameter', 'Language', 'Frequency', 'Confidence', 'References']

    def row(self, ctx, req, item):
        res = super(Values, self).row(ctx, req, item)
        for obj in [item.valueset.parameter, item.valueset.language]:
            res.append(hyperlink(req.resource_url(obj), obj.__unicode__()))
        res.extend([item.frequency or '', item.confidence or ''])
        res.append(';'.join(filter(
            None, [r.source.name for r in item.valueset.references if r.source])))
        return res


class Sentences(ExcelAdapter):

    """Represent table of Sentence instances as excel sheet (maximal %d rows)."""

    __doc__ %= QUERY_LIMIT

    def header(self, ctx, req):
        return ['ID', 'Text', 'Analyzed', 'Gloss', 'Translation', 'Language']

    def row(self, ctx, req, item):
        return [
            hyperlink(req.resource_url(item), item.id),
            item.name,
            item.analyzed,
            item.gloss,
            item.description,
            hyperlink(req.resource_url(item.language), item.language.name),
        ]



"""Functionality for alembic scripts.

This module provides

- basic crud functionality within alembic migration scripts,
- advanced helpers for special tasks, like merging sources.

.. note::

    Using the functionality provided in this module is not possible for Alembic scripts
    supposed to be run in
    `offline mode <http://alembic.readthedocs.org/en/latest/tutorial.html\
    #generating-sql-scripts-a-k-a-offline-mode>`_.
"""

from sqlalchemy.sql import select as base_select, func
from clld.db.models import common


def _with_where_clause(table, stmt, **where):
    for key, value in where.items():
        stmt = stmt.where(getattr(table.c, key) == value)
    return stmt


class Connection(object):

    """A wrapper around an SQLAlchemy connection.

    This wrapper provides the convenience of allowing typical CRUD operations to be
    called passing model classes.

    Additionally, it implements more complicated clld domain specific database
    operations.

    A ``Connection`` will typically be instantiated in an Alembic migration script as
    follows:

    .. code-block:: python

        from alembic import op
        conn = Connection(op.get_bind())
    """

    def __init__(self, conn):
        """Initialize.

        :param conn:
            SQLAlchemy connection object, i.e. anything with a ``execute`` method; so in
            particular a session qualifies as well as an engine, i.e. what ``get_bind()``
            returns.
        """
        self._conn = conn

    def execute(self, *args, **kw):
        """Provide access to the underlying connection's ``execute`` method."""
        return self._conn.execute(*args, **kw)

    #
    # CRUD operations:
    #
    def select(self, model, **where):
        """Run a select statement and return a ResultProxy."""
        table = model.__table__
        return self.execute(_with_where_clause(table, base_select([table]), **where))

    def all(self, model, **where):
        """return all results of a select statement."""
        return self.select(model, **where).fetchall()

    def first(self, model, **where):
        """return first result of a select statement or ``None``."""
        return self.select(model, **where).fetchone()

    def get(self, model, pk):
        """return row specified by primary key."""
        return self.first(model, pk=pk)

    def pk(self, model, id_, attr='id'):
        """Get the primary key of an object specified by a unique property.

        :param model: model class.
        :param id_: Value to be used when filtering.
        :param attr: Column to be used for filtering.
        :return: primary key of (first) matching row.
        """
        res = self.first(model, **{attr: id_})
        if res:
            return res.pk

    def insert(self, model, **values):
        """Run an insert statement.

        :return: primary key of the inserted row.
        """
        for k, v in [
            ('version', 1),
            ('created', func.now()),
            ('updated', func.now()),
            ('active', True)
        ]:
            if hasattr(model.__table__.c, k):
                values.setdefault(k, v)

        res = self.execute(model.__table__.insert().values(**values))
        return res.inserted_primary_key[0]

    def update(self, model, values, **where):
        """Run an update statement."""
        if not isinstance(values, dict):
            values = dict(values)
        if hasattr(model.__table__.c, 'updated'):
            values.setdefault('updated', func.now())
        table = model.__table__
        self.execute(_with_where_clause(table, table.update(), **where).values(**values))

    def delete(self, model, **where):
        """Run a delete statement."""
        self.execute(
            _with_where_clause(model.__table__, model.__table__.delete(), **where))

    #
    # domain specific operations:
    #
    def set_glottocode(self, lid, gc, gcid=None):
        """assign a unique glottocode to a language.

        i.e. alternative glottocodes will be deleted.

        :param lid: ``id`` of the language.
        :param gc: Glottocode to be assigned.
        :param gcid:
            ``id`` of the ``Identifier`` instance if one has to be created;
            defaults to ``gc``.
        """
        lpk = self.pk(common.Language, lid)
        gctype = common.IdentifierType.glottolog.value

        done = False
        lis = self.all(common.LanguageIdentifier, language_pk=lpk)
        for li in lis:
            i = self.get(common.Identifier, li.identifier_pk)
            if i.type == gctype:
                if i.name == gc:
                    done = True
                else:
                    self.delete(common.LanguageIdentifier, pk=li.pk)

        if not done:
            # create a new relation
            i = self.first(common.Identifier, name=gc, type=gctype)
            if i:
                ipk = i.pk
            else:
                ipk = self.insert(common.Identifier, id=gcid or gc, name=gc, type=gctype)
            return self.insert(
                common.LanguageIdentifier, language_pk=lpk, identifier_pk=ipk)

    def replace(self, model, from_id, to_id):  # pragma: no cover
        self.insert(
            common.Config, key=common.Config.replacement_key(model, from_id), value=to_id)


def merge_sources(conn, from_id, to_id, *fields):  # pragma: no cover
    if not isinstance(conn, Connection):
        conn = Connection(conn)
    # resolve id to pk
    fpk = conn.pk(common.Source, from_id)
    tpk = conn.pk(common.Source, to_id)

    if fields:
        conn.execute("""\
UPDATE source
SET %s
FROM (SELECT %s FROM source WHERE pk = %s) AS old
WHERE pk = %s""" % (', '.join('{0}=old.{0}'.format(f) for f in fields),
                    ', '.join(fields),
                    fpk,
                    tpk))

    # update relationships
    for row in conn.all(common.LanguageSource, source_pk=fpk):
        lpk = row['language_pk']
        if conn.first(common.LanguageSource, language_pk=lpk, source_pk=tpk):
            conn.delete(common.LanguageSource, source_pk=fpk)
        else:
            conn.update(common.LanguageSource, dict(source_pk=tpk), source_pk=fpk)

    for model in [
        common.SentenceReference, common.ContributionReference, common.ValueSetReference
    ]:
        conn.update(model, dict(source_pk=tpk), source_pk=fpk)

    # add replacement record
    conn.replace(common.Source, from_id, to_id)

    # remove source
    conn.delete(common.Source, id=from_id)



"""View logic for clld default views."""
from json import dumps
import re
from functools import partial
from datetime import datetime
from time import mktime

import requests
from requests.exceptions import Timeout
import feedparser
from pyramid.response import Response
import pyramid.httpexceptions
from pyramid.interfaces import IRoutesMapper
from pyramid.renderers import render, render_to_response

from clld.util import summary
from clld.interfaces import IRepresentation, IIndex, IMetadata
from clld.web.adapters import get_adapter, get_adapters
from clld.web.adapters.csv import CsvAdapter, CsvmJsonAdapter
from clld.web.util.multiselect import MultiSelect
from clld.db.models.common import Combination
from clld.web.maps import CombinedMap


def xpartial(func, *args, **kw):
    """Augment partial to make it possible to register partials as view callables.

    >>> f = xpartial(lambda x, y: x + y, 1)
    >>> assert f(1) == 2
    """
    res = partial(func, *args, **kw)
    res.__module__ = func.__module__
    return res


def redirect(cls, location, ctx, req=None):
    """Can be used with xpartial to register views that simply redirect.

    >>> view = xpartial(pyramid.httpexceptions.HTTPFound, lambda req: req.route_url('.'))
    """
    # workaround for https://github.com/Pylons/pyramid/issues/1428
    req = req or ctx
    if callable(location):
        location = location(req)
    raise cls(location=location)


def gone(ctx, req):
    """view callable."""
    raise pyramid.httpexceptions.HTTPGone()


def view(interface, ctx, req, getadapters=False):
    """Render a resource as pyramid response.

    Using the most appropriate adapter for the accept header sent.

    :param getadapters: If True, the adapter used to render the response and the list of\
    all available adapters for the context are returned as well.
    """
    #
    # if req.matched_route.name.endswith('_alt') -> add rel="canonical" header
    # else: add rel="alternate" header
    #
    adapter, adapters = get_adapter(
        interface, ctx, req, ext=req.matchdict and req.matchdict.get('ext'), getall=True)
    if not adapter:
        raise pyramid.httpexceptions.HTTPNotAcceptable()
    res = adapter.render_to_response(ctx, req)
    if getadapters:
        return res, adapter, adapters
    return res  # pragma: no cover


def _add_link_header(response, url, adapter=None, rel="canonical", mimetype="text/html"):
    if adapter:
        rel = adapter.rel
        mimetype = adapter.send_mimetype or adapter.mimetype
    if mimetype and rel:
        response.headerlist.append(
            ('Link', '<%s>; rel="%s"; type="%s"' % tuple(map(str, [url, rel, mimetype]))))


def index_view(ctx, req):
    if req.is_xhr and 'sEcho' in req.params:
        return datatable_xhr_view(ctx, req)
    res, current, adapters = view(IIndex, ctx, req, getadapters=True)
    if req.matched_route:
        if req.matched_route.name.endswith('_alt'):
            # add the canonical link:
            _add_link_header(res, req.route_url(req.matched_route.name[:-4]))
            if current.mimetype == CsvAdapter.mimetype:
                # If serving csv add link header to make metadata discoverable. See
                # http://www.w3.org/TR/2015/CR-tabular-data-model-20150716/#link-header
                csvm = [a for a in adapters if a.extension == CsvmJsonAdapter.extension]
                if csvm:
                    a = csvm[0]
                    _add_link_header(
                        res,
                        req.route_url(req.matched_route.name, ext=a.extension),
                        adapter=a)
        else:
            for a in [_a for _a in adapters if _a.rel and _a.extension]:
                _add_link_header(
                    res,
                    req.route_url(req.matched_route.name + '_alt', ext=a.extension),
                    adapter=a)
    return res


def resource_view(ctx, req):
    res, _, adapters = view(IRepresentation, ctx, req, getadapters=True)
    if req.matched_route:
        if req.matched_route.name.endswith('_alt'):
            _add_link_header(res, req.resource_url(ctx))
        else:
            for a in adapters:
                if a.rel and a.extension:
                    _add_link_header(
                        res, req.resource_url(ctx, ext=a.extension), adapter=a)
    return res


def datatable_xhr_view(ctx, req):
    # call get_query, thereby - as side effect - making sure, the counts are set.
    items = ctx.get_query()
    if hasattr(ctx, 'row_class'):
        data = []
        for item in items:
            _d = {'DT_RowId': 'row_%s' % item.pk}
            row_class = ctx.row_class(item)
            if row_class:
                _d['DT_RowClass'] = row_class
            for i, col in enumerate(ctx.cols):
                _d[str(i)] = col.format(item)
            data.append(_d)
    else:
        data = [[col.format(item) for col in ctx.cols] for item in items]

    # sEcho parameter.
    # Note that it strongly recommended for security reasons that you 'cast' this
    # parameter to an integer in order to prevent Cross Site Scripting (XSS) attacks.
    try:
        echo = int(req.params['sEcho'])
    except ValueError:
        echo = 1

    res = {
        "aaData": data,
        "sEcho": str(echo),
        "iTotalRecords": ctx.count_all,
        "iTotalDisplayRecords": ctx.count_filtered,
    }
    return render_to_response('json', res, request=req)


def js(req):
    param_pattern = re.compile('\{(?P<name>[a-z]+)(\:[^\}]+)?\}')

    res = [
        "CLLD.base_url = %s;" % dumps(req.application_url),
        "CLLD.query_params = %s;" % dumps(req.query_params),
    ]
    for route in req.registry.getUtility(IRoutesMapper).get_routes():
        pattern = param_pattern.sub(lambda m: '{%s}' % m.group('name'), route.pattern)
        res.append('CLLD.routes[%s] = %s;' % tuple(map(dumps, [route.name, pattern])))
    return Response('\n'.join(res), content_type="text/javascript")


def select_combination(ctx, req):
    if 'parameters' in req.params:
        ids = req.params.getall('parameters')
        if len(ids) > 1:
            id_ = Combination.delimiter.join(ids)
        else:
            id_ = Combination.delimiter.join(req.params['parameters'].split(','))
        raise pyramid.httpexceptions.HTTPFound(req.route_url('combination', id=id_))
    raise pyramid.httpexceptions.HTTPNotFound


def _raise(req):
    """view callable to test error reporting in running apps."""
    raise ValueError('test')


def _ping(req):
    """view to test aliveness of apps."""
    return {'status': 'ok'}


def unapi(req):
    """View callable implementing the server side of the unAPI spec."""
    id_ = req.params.get('id')
    if id_:
        obj = req.ctx_for_url(id_)
        if not obj:
            return pyramid.httpexceptions.HTTPNotFound()

    format_ = req.params.get('format')
    if not format_:
        kw = {'content_type': 'application/xml'}
        if id_:
            formats = [a for n, a in get_adapters(IMetadata, obj, req)]
            kw['status'] = '300 Multiple Choices'
        else:
            formats = []
        body = render(
            'unapiformats.mako',
            {'formats': formats, 'identifier': id_},
            request=req)
        return Response(body, **kw)

    if not id_:
        return pyramid.httpexceptions.HTTPNotFound()

    adapter = None
    for _n, _a in get_adapters(IMetadata, obj, req):
        if _a.unapi_name == format_:
            adapter = _a
            break
    if not adapter:
        return pyramid.httpexceptions.HTTPNotAcceptable()
    return pyramid.httpexceptions.HTTPFound(
        location=req.resource_url(obj, ext=adapter.extension))


#
# The following code implements a view to show a map with parameters from distinct
# datasets. It may be used by CrossGram at some point.
#
class ParameterMultiSelect(MultiSelect):  # pragma: no cover

    """Experimental."""

    def __init__(self, req, name, eid, collection=None, url=None, selected=None):
        MultiSelect.__init__(self, req, name, eid, url='x')
        self.data = []
        for app, rm in req.registry.settings.get('clld.parameters', {}).items():
            for param in rm['resources']:
                self.data.append({
                    'id': '%s-%s' % (app, param['id']),
                    'text': '%s %s: %s' % (app, param['id'], param['name'])})
        self._datadict = dict((d['id'], d) for d in self.data)

    def format_result(self, obj):
        return obj

    def get_urls(self):
        for i, param in enumerate(self.req.params.get('parameters', '').split(',')):
            if param in self._datadict:
                if self.selected is None:
                    self.selected = []
                self.selected.append(self._datadict[param])
                app, pid = param.split('-', 1)
                rm = self.req.registry.settings['clld.parameters'][app]
                yield (app, pid, rm['properties']['uri_template'].format(id=pid))

    def get_default_options(self):
        return {
            'placeholder': "Select parameter",
            'width': 'off',
            'class': 'span6',
            'multiple': True,
            'data': self.data}


def combined(ctx, req):  # pragma: no cover
    res = {'map': None, 'select': ParameterMultiSelect(req, 'parameters', 'parameters')}
    urls = list(res['select'].get_urls())
    if urls:
        res['map'] = CombinedMap(urls, req)
    return res


def atom_feed(request, feed_url):
    """
    Proxy feeds so they can be accessed via XHR requests.

    We also convert RSS to ATOM so that the javascript Feed component can read them.
    """
    ctx = {'url': feed_url, 'title': None, 'entries': []}
    try:
        res = requests.get(ctx['url'], timeout=(3.05, 1))
    except Timeout:
        res = None
    if res and res.status_code == 200:
        d = feedparser.parse(res.content.strip())
        ctx['title'] = getattr(d.feed, 'title', None)
        for e in d.entries:
            ctx['entries'].append(dict(
                title=e.title,
                link=e.link,
                updated=datetime.fromtimestamp(mktime(e.published_parsed)).isoformat(),
                summary=summary(e.description)))
    response = render_to_response('atom_feed.mako', ctx, request=request)
    response.content_type = 'application/atom+xml'
    return response



from __future__ import unicode_literals, print_function, division, absolute_import

from sqlalchemy import Column, Integer, Boolean, Date, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld.db.versioned import Versioned
from clld import interfaces

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    HasSourceNotNullMixin, Contributor)

__all__ = ('Contribution', 'ContributionReference', 'ContributionContributor')


class Contribution_data(Base, Versioned, DataMixin):
    pass


class Contribution_files(Base, Versioned, FilesMixin):
    pass


@implementer(interfaces.IContribution)
class Contribution(Base,
                   PolymorphicBaseMixin,
                   Versioned,
                   IdNameDescriptionMixin,
                   HasDataMixin,
                   HasFilesMixin):

    """A set of data contributed within the same context by the same contributors."""

    __table_args__ = (UniqueConstraint('name'),)

    date = Column(Date)

    @property
    def primary_contributors(self):
        return [assoc.contributor for assoc in
                sorted(self.contributor_assocs,
                       key=lambda a: (a.ord, a.contributor.id)) if assoc.primary]

    @property
    def secondary_contributors(self):
        return [assoc.contributor for assoc in
                sorted(self.contributor_assocs,
                       key=lambda a: (a.ord, a.contributor.id)) if not assoc.primary]

    def formatted_contributors(self):
        contribs = [' and '.join(c.name for c in self.primary_contributors)]
        if self.secondary_contributors:
            contribs.append(' and '.join(c.name for c in self.secondary_contributors))
        return ' with '.join(contribs)


class ContributionReference(Base, Versioned, HasSourceNotNullMixin):

    """Association table."""

    __table_args__ = (
        UniqueConstraint('contribution_pk', 'source_pk', 'description'),
    )

    contribution_pk = Column(Integer, ForeignKey('contribution.pk'), nullable=False)
    contribution = relationship(Contribution, innerjoin=True, backref="references")


class ContributionContributor(Base, PolymorphicBaseMixin, Versioned):

    """Many-to-many association between contributors and contributions."""

    __table_args__ = (UniqueConstraint('contribution_pk', 'contributor_pk'),)

    contribution_pk = Column(Integer, ForeignKey('contribution.pk'), nullable=False)
    contributor_pk = Column(Integer, ForeignKey('contributor.pk'), nullable=False)

    # contributors are ordered.
    ord = Column(Integer, default=1)

    # we distinguish between primary and secondary (a.k.a. 'with ...') contributors.
    primary = Column(Boolean, default=True)

    contribution = relationship(
        Contribution, innerjoin=True, backref='contributor_assocs')
    contributor = relationship(
        Contributor, innerjoin=True, lazy=False, backref='contribution_assocs')



"""
Functionality to be called from command line scripts is provided in this module.

.. note:

    The various scripts are installed when running `python setup.py develop|install`,
    following the specification in `setup.py`.

.. seealso: http://www.scotttorborg.com/python-packaging/command-line-scripts.html\
#the-console-scripts-entry-point
"""
from __future__ import unicode_literals, division, absolute_import, print_function
import os
import argparse

import transaction

from clld.interfaces import IDownload
from clld.scripts.util import parsed_args, gbs_func
from clld.scripts.internetarchive import ia_func


def internetarchive(**kw):  # pragma: no cover
    """
    Add information about availability on Internet Archive to Source objects.
    """
    add_args = [(("command",), dict(help="download|verify|update"))]
    kw.setdefault('description', internetarchive.__doc__)
    args = parsed_args(*add_args, **kw)
    with transaction.manager:
        ia_func(args.command, args, kw.get('sources'))


def create_downloads(**kw):  # pragma: no cover
    """
    Create all registered downloads (locally).
    """
    args = parsed_args(bootstrap=True, description=create_downloads.__doc__)
    for name, download in args.env['registry'].getUtilitiesFor(IDownload):
        args.log.info('creating download %s' % name)
        download.create(args.env['request'])


def google_books(**kw):  # pragma: no cover
    add_args = [
        (("command",), dict(help="download|verify|update|cleanup")),
        (("--api-key",), dict(default=kw.get('key', os.environ.get('GBS_API_KEY')))),
    ]

    args = parsed_args(*add_args, **kw)
    if args.command == 'download' and not args.api_key:
        raise argparse.ArgumentError(None, 'no API key found for download')

    with transaction.manager:
        gbs_func(args.command, args, kw.get('sources'))



"""
python interface to bibutils.

.. seealso:: http://sourceforge.net/p/bibutils/home/Bibutils/
"""
from subprocess import Popen, PIPE

from clldutils.misc import encoded


FORMATS = ['bib', 'end', 'ris']


def pipe(cmd, input_):
    proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
    return proc.communicate(encoded(input_))[0]


def convert(string, from_, to_=None):
    assert from_ in FORMATS and (to_ is None or to_ in FORMATS)

    res = pipe('%s2xml -i utf8 -o utf8' % from_, encoded(string))
    if to_:
        res = pipe('xml2%s -i utf8 -o utf8' % to_, res)
    return res.decode('utf8')



from __future__ import unicode_literals
from collections import OrderedDict, defaultdict
import shutil

import transaction
from zope.interface import implementer
from sqlalchemy.orm import joinedload_all, joinedload
from pycldf import dataset
from pycldf import sources
from clldutils.path import TemporaryDirectory
from csvw.metadata import URITemplate, Column

from clld.web.adapters.download import Download
from clld.db.meta import DBSession
from clld.db.models.common import (
    ValueSet, Value, Parameter, Language, Contribution, Source, Sentence,
    ContributionContributor, DomainElement,
)
from clld.web.util.helpers import text_citation, get_url_template
from clld.interfaces import ICldfConfig


def url_template(req, route, id_name):
    return URITemplate(get_url_template(
        req, route, relative=False, variable_map={'id': id_name}))


def source2source(req, source):
    """Harmonize the different Source implementations in clld and pycldf."""
    bibrecord = source.bibtex()
    fields = OrderedDict({'%s_url' % req.dataset.id: req.resource_url(source)})
    for key, value in bibrecord.items():
        fields[key] = '; '.join(value) if isinstance(value, list) else value
    return sources.Source(
        getattr(bibrecord.genre, 'value', bibrecord.genre) if bibrecord.genre else 'misc',
        source.id,
        **fields)


def iterrefs(obj):
    def _desc(d):
        return '[%s]' % d.replace(';', '.').replace('[', '{').replace(']', '}') \
            if d else ''

    # For backwards compatibility:
    if not hasattr(obj, 'references') and hasattr(obj, 'valueset'):
        obj = obj.valueset

    for r in obj.references:
        if r.source_pk:
            yield r.source_pk, _desc(r.description)


@implementer(ICldfConfig)
class CldfConfig(object):
    module = 'Wordlist'
    pk2id = defaultdict(dict)

    def custom_schema(self, req, ds):
        return

    def query(self, model):
        q = DBSession.query(model).order_by(model.pk)
        if model == Contribution:
            return q.options(joinedload_all(
                Contribution.contributor_assocs,
                ContributionContributor.contributor))
        if model == Sentence:
            return q.options(joinedload(Sentence.language))
        if model == DomainElement:
            return q.order_by(None).order_by(model.parameter_pk, model.number, model.pk)
        if model == Value:
            return q.join(ValueSet)\
                .order_by(None)\
                .order_by(
                    ValueSet.parameter_pk,
                    ValueSet.language_pk,
                    ValueSet.contribution_pk,
                    Value.pk)\
                .options(
                    joinedload_all(Value.valueset, ValueSet.references),
                    joinedload(Value.domainelement))
        return q

    def convert(self, model, item, req):
        self.pk2id[model.__name__][item.pk] = item.id
        if model == Parameter:
            return {'ID': item.id, 'Name': item.name, 'Description': item.description}
        if model == DomainElement:
            return {
                'ID': item.id,
                'Name': item.name,
                'Description': item.description,
                'Parameter_ID': self.pk2id['Parameter'][item.parameter_pk],
                'Number': item.number,
            }
        if model == Language:
            return {
                'ID': item.id,
                'Name': item.name,
                'Glottocode': item.glottocode,
                'ISO639P3code': item.iso_code,
                'Latitude': item.latitude,
                'Longitude': item.longitude,
            }
        if model == Contribution:
            return {
                'ID': item.id,
                'Name': item.name,
                'Description': item.description,
                'Contributors': item.formatted_contributors(),
            }
        if model == Sentence:
            return {
                'ID': item.id,
                'Language_ID': self.pk2id['Language'][item.language_pk],
                'Primary_Text': item.name,
                'Analyzed_Word': item.analyzed.split('\t') if item.analyzed else [],
                'Gloss': item.gloss.split('\t') if item.gloss else [],
                'Translated_Text': item.description,
                'Comment': item.comment,
            }
        if model == Source:
            return source2source(req, item)
        if model == Value:
            res = {
                'ID': item.id,
                'Language_ID': self.pk2id['Language'][item.valueset.language_pk],
                'Parameter_ID': self.pk2id['Parameter'][item.valueset.parameter_pk],
                'Contribution_ID': self.pk2id['Contribution'][item.valueset.contribution_pk],
                'Value': (item.domainelement.name if item.domainelement else item.name) or '-',
                'Source': [
                    '{0}{1}'.format(self.pk2id['Source'][spk], d) for spk, d in iterrefs(item)],
            }
            if item.domainelement_pk:
                res['Code_ID'] = self.pk2id['DomainElement'][item.domainelement_pk]
            if self.module == 'Wordlist':
                res['Form'] = res['Value']
            return res
        raise Value(model)  # pragma: no cover

    def custom_tabledata(self, req, tabledata):
        return tabledata


class CldfDownload(Download):
    ext = 'cldf'
    description = "Dataset in CLDF"

    def create(self, req, filename=None, verbose=True, outfile=None):
        cldf_cfg = req.registry.getUtility(ICldfConfig)

        with TemporaryDirectory() as tmpd:
            cls = getattr(dataset, cldf_cfg.module)
            ds = cls.in_dir(tmpd)
            ds.properties['dc:bibliographicCitation '] = text_citation(req, req.dataset)
            ds.properties['dc:publisher'] = '%s, %s' % (
                req.dataset.publisher_name, req.dataset.publisher_place)
            ds.properties['dc:license'] = req.dataset.license
            ds.properties['dc:issued'] = req.dataset.published.isoformat()
            ds.properties['dc:title'] = req.dataset.name
            ds.properties['dc:creator'] = req.dataset.formatted_editors()
            ds.properties['dc:identifier'] = req.resource_url(req.dataset)
            ds.properties['dcat:accessURL'] = req.route_url('download')
            if DBSession.query(Sentence).count():
                ds.add_component('ExampleTable')
            if DBSession.query(DomainElement).count():
                ds.add_component('CodeTable', {'name': 'Number', 'datatype': 'integer'})
            ds.add_component('ParameterTable')
            ds.add_component('LanguageTable')
            ds.add_table('contributions.csv', 'ID', 'Name', 'Description', 'Contributors')
            ds.add_columns(ds.primary_table, Column.fromvalue(
                {
                    'name': 'Contribution_ID',
                    'datatype': 'string',
                    'valueUrl': url_template(req, 'contribution', 'contribution').uri,
                }))
            ds.add_foreign_key(
                ds.primary_table, 'Contribution_ID', 'contributions.csv', 'ID')
            ds['LanguageTable'].aboutUrl = url_template(req, 'language', 'ID')
            ds['ParameterTable'].aboutUrl = url_template(req, 'parameter', 'ID')
            ds[ds.primary_table].aboutUrl = url_template(req, 'value', 'ID')

            cldf_cfg.custom_schema(req, ds)

            for src in cldf_cfg.query(Source):
                ds.sources.add(cldf_cfg.convert(Source, src, req))
            fname = outfile or self.abspath(req)

            transaction.abort()

            tabledata = defaultdict(list)
            for table, model in [
                ('ParameterTable', Parameter),
                ('CodeTable', DomainElement),
                ('LanguageTable', Language),
                ('ExampleTable', Sentence),
                ('contributions.csv', Contribution),
                (ds.primary_table, Value),
            ]:
                if verbose:
                    print('exporting {0} ...'.format(model))
                transaction.begin()
                for item in cldf_cfg.query(model):
                    tabledata[table].append(cldf_cfg.convert(model, item, req))
                transaction.abort()
                if verbose:
                    print('... done')

            transaction.begin()
            ds.write(**cldf_cfg.custom_tabledata(req, tabledata))
            ds.validate()

            shutil.make_archive(
                fname.parent.joinpath(fname.stem).as_posix(), 'zip', tmpd.as_posix())



from setuptools import setup, find_packages

install_requires = [
    'Babel',
    'csvw~=1.0',
    'clldutils~=2.4',
    'pycldf>=1.0.7',
    'setuptools>=25',
    'pyramid>=1.6',
    'pyramid_mako>=1.0',
    'pyramid_tm',
    'SQLAlchemy>=1.0.6',
    'purl>=0.5',
    'pytz',
    'zope.sqlalchemy',
    'six>=1.7.3',  # webassets needs add_metaclass!
    'alembic>=0.7.1',
    'webassets',
    'yuicompressor',
    'markupsafe',
    'requests>=2.4.3',  # we use the support for connect timeouts introduced in 2.4.0
    'rdflib>=4.1.1',  # rdflib 4.1.0 requires html5lib==0.95
    'colander',
    'python-dateutil',
    'paginate',
    'xlwt',
    'webhelpers2>=2.0',
    'nameparser',
    'feedparser',
    'waitress',
]


setup(
    name='clld',
    version='4.4.0',
    description=(
        'Python library supporting the development of cross-linguistic databases'),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
    ],
    keywords='web pyramid LRL Linguistics',
    author="Robert Forkel, MPI SHH",
    author_email="forkel@shh.mpg.de",
    url="http://clld.org",
    license="Apache Software License",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
    install_requires=install_requires,
    extras_require={
        'dev': [
            'waitress',
            'pyramid_debugtoolbar',
            'tox',
            'flake8',
            'wheel',
            'twine',
        ],
        'test': [
            'xlrd',
            'mock',
            'pytest>=3.1',
            'pytest-clld',
            'pytest-mock',
            'pytest-cov',
            'coverage>=4.2',
            'zope.component>=3.11.0',
        ],
        'docs': [
            'Sphinx',
            'docutils',
            'repoze.sphinx.autointerface',
        ],
    },
    message_extractors={'src/clld': [
        ('**.py', 'python', None),
        ('**.mako', 'mako', {'encoding': 'utf8'}),
        ('web/static/**', 'ignore', None)]},
    entry_points="""\
        [pyramid.scaffold]
        clld_app=clld.scaffolds:ClldAppTemplate
        [console_scripts]
        clld-google-books = clld.scripts.cli:google_books
        clld-internetarchive = clld.scripts.cli:internetarchive
        clld-create-downloads = clld.scripts.cli:create_downloads
    """
)



from __future__ import unicode_literals, print_function, division, absolute_import

from sqlalchemy import Column, Integer, String, Unicode, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr

from zope.interface import implementer
from six import iteritems

from clld.db.meta import Base, PolymorphicBaseMixin
from clld.db.versioned import Versioned
from clld import interfaces
from clld.lib import bibtex
from clld.lib import coins
from clld.web.util.htmllib import HTML

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    Language, LanguageSource)

__all__ = ('Source', 'HasSourceMixin', 'HasSourceNotNullMixin')


class Source_data(Base, Versioned, DataMixin):
    pass


class Source_files(Base, Versioned, FilesMixin):
    pass


@implementer(interfaces.ISource)
class Source(Base,
             PolymorphicBaseMixin,
             Versioned,
             IdNameDescriptionMixin,
             HasDataMixin,
             HasFilesMixin):

    """A bibliographic record, cited as source for some statement."""

    glottolog_id = Column(String)
    google_book_search_id = Column(String)

    #
    # BibTeX fields:
    #
    bibtex_type = Column(bibtex.EntryType.db_type())
    author = Column(Unicode)
    year = Column(Unicode)
    title = Column(Unicode)
    type = Column(Unicode)
    booktitle = Column(Unicode)
    editor = Column(Unicode)
    pages = Column(Unicode)
    edition = Column(Unicode)
    journal = Column(Unicode)
    school = Column(Unicode)
    address = Column(Unicode)
    url = Column(Unicode)
    note = Column(Unicode)
    number = Column(Unicode)
    series = Column(Unicode)
    volume = Column(Unicode)
    publisher = Column(Unicode)
    organization = Column(Unicode)
    chapter = Column(Unicode)
    howpublished = Column(Unicode)

    # typed information we might want to use for searching or sorting:
    year_int = Column(Integer)
    startpage_int = Column(Integer)
    pages_int = Column(Integer)

    languages = relationship(
        Language, backref='sources', secondary=LanguageSource.__table__)

    @property
    def gbs_identifier(self):
        if not self.jsondata.get('gbs'):
            return
        if not self.jsondata['gbs']['volumeInfo'].get('industryIdentifiers'):
            return
        id_ = None
        for identifier in self.jsondata['gbs']['volumeInfo']['industryIdentifiers']:
            # prefer ISBN_13 over ISBN_10 over anything else
            if identifier['type'] == 'ISBN_13':
                id_ = 'ISBN:' + identifier['identifier']
            if identifier['type'] == 'ISBN_10' and not id_:
                id_ = 'ISBN:' + identifier['identifier']
        if not id_:
            # grab the last one in the list (most probably the only one!)
            id_ = identifier['identifier']
        return id_

    def __bibtex__(self):
        return {}

    def bibtex(self, exclude={'gbs', 'glottolog_ref_id'}):
        kw = {k: v for k, v in iteritems(self.jsondata) if k not in exclude}
        kw.update(self.__bibtex__())
        return bibtex.Record.from_object(self, **kw)

    def coins(self, req):
        return HTML.span(
            ' ',
            **coins.ContextObject.from_bibtex(
                req.dataset.name, self.bibtex()).span_attrs()
        )


#
# Several objects can be linked to sources, i.e. they can have references.
#
class HasSourceMixin(object):
    key = Column(Unicode)  # the citation key, specific (and unique) within a contribution
    description = Column(Unicode)  # e.g. page numbers.

    @declared_attr
    def source_pk(cls):  # pragma: no cover
        return Column(Integer, ForeignKey('source.pk'))

    @declared_attr
    def source(cls):  # pragma: no cover
        return relationship(Source, backref=cls.__name__.lower() + 's')


class HasSourceNotNullMixin(HasSourceMixin):

    @declared_attr
    def source_pk(cls):
        return Column(Integer, ForeignKey('source.pk'), nullable=False)

    @declared_attr
    def source(cls):
        return relationship(Source, innerjoin=True, backref=cls.__name__.lower() + 's')



"""Functionality to manage icons for map markers."""
from __future__ import unicode_literals, print_function, division, absolute_import
from itertools import product, chain

from zope.interface import implementer
from clld.interfaces import IIcon, IMapMarker


SHAPES = [
    "c",  # circle
    "s",  # square
    "t",  # triangle (pyramid)
    "f",  # inverted pyramid
    "d",  # diamond
]

COLORS = [
    '000000',
    '0000dd',
    '0000ff',
    '009900',
    '00ff00',
    '00ffff',
    '4d6cee',
    '66ff33',
    '990099',
    '9999ff',
    '99ffff',
    'a0fb75',
    'aa0000',
    'cb9a34',
    'cccccc',
    'd22257',
    'dd0000',
    'e8e8e8',
    'ed9c07',
    'efe305',
    'f38847',
    'f3ffb0',
    'fe3856',
    'ff0000',
    'ff00ff',
    'ff4400',
    'ff6600',
    'ff66ff',
    'ffcc00',
    'ffff00',
    'ffffcc',
    'ffffff',
]

# the following colors show enough mutual contrast to be easily distinguised:
PREFERED_COLORS = [
    '0000dd',
    '009900',
    '990099',
    'dd0000',
    'ffff00',
    'ffffff',
    '00ff00',
    '00ffff',
    'cccccc',
    'ff6600',
]
SECONDARY_COLORS = [c for c in COLORS if c not in PREFERED_COLORS]


@implementer(IIcon)
class Icon(object):

    """Default implementation of IIcon: Icons are static image files."""

    def __init__(self, name):
        self.name = name

    @property
    def asset_spec(self):
        return 'clld:web/static/icons/%s.png' % self.name

    def url(self, req):
        return req.static_url(self.asset_spec)


#: a list of all available icons:
ICONS = [Icon('%s%s' % (s, c)) for s in SHAPES for c in COLORS]

#: a dictionary mapping icon names to icon objects:
ICON_MAP = {icon.name: icon for icon in ICONS}

#: a list of icons ordered by preference:
ORDERED_ICONS = [ICON_MAP[s + c] for s, c in chain(
                 product(SHAPES, PREFERED_COLORS),
                 product(SHAPES, SECONDARY_COLORS))]


@implementer(IMapMarker)
class MapMarker(object):

    """The default map marker is an orange circle."""

    def get_icon(self, ctx, req):
        return 'cff6600'

    def __call__(self, ctx, req):
        icon = self.get_icon(ctx, req) or 'cff6600'
        return req.registry.getUtility(IIcon, icon).url(req)



"""
Serialize clld objects as csv.

We also provide metadata following the guidelines of

- http://www.w3.org/TR/tabular-data-model/ version
  http://www.w3.org/TR/2015/CR-tabular-data-model-20150716/
- http://www.w3.org/TR/tabular-metadata/ version
  http://www.w3.org/TR/2015/CR-tabular-metadata-20150716/
"""
from __future__ import unicode_literals, print_function, division, absolute_import
from itertools import chain

from sqlalchemy import types, Column
from sqlalchemy.inspection import inspect
from pyramid.renderers import render as pyramid_render
from csvw.dsv import UnicodeWriter

from clld.web.adapters.base import Index

QUERY_LIMIT = 2000


class CsvAdapter(Index):

    """Represent tables as csv files (maximal %d rows)."""

    __doc__ %= QUERY_LIMIT

    extension = 'csv'
    mimetype = 'text/csv'
    content_type_params = dict(header='present')

    def render(self, ctx, req):
        with UnicodeWriter() as writer:
            rows = iter(ctx.get_query(limit=QUERY_LIMIT))
            first = next(rows, None)
            if first is not None:
                cols = first.csv_head()
                writer.writerow(cols)
                for item in chain([first], rows):
                    writer.writerow(item.to_csv(ctx=ctx, req=req, cols=cols))
            return writer.read()

    def render_to_response(self, ctx, req):
        res = super(CsvAdapter, self).render_to_response(ctx, req)
        res.content_disposition = 'attachment; filename="%s.csv"' % repr(ctx)
        return res


class CsvmJsonAdapter(Index):

    """Metadata for tabular data serialited as csv.

    .. seealso:: http://www.w3.org/TR/tabular-metadata/
    """

    extension = 'csv-metadata.json'
    mimetype = 'application/csvm+json'
    send_mimetype = 'application/csvm+json'
    rel = 'describedby'

    type_map = [
        (types.Integer, 'http://www.w3.org/2001/XMLSchema#int'),
        (types.Float, 'http://www.w3.org/2001/XMLSchema#float'),
        (types.Boolean, 'http://www.w3.org/2001/XMLSchema#boolean'),
        (types.DateTime, 'http://www.w3.org/2001/XMLSchema#dateTime'),
        (types.Date, 'http://www.w3.org/2001/XMLSchema#date'),
        (types.Unicode, 'http://www.w3.org/2001/XMLSchema#string'),
        (types.String, 'http://www.w3.org/2001/XMLSchema#string'),
    ]

    @classmethod
    def csvm_basic_doc(cls, req, **kw):
        doc = {
            "@context": ["http://www.w3.org/ns/csvw", {"@language": "en"}],
            "dc:publisher": {
                "schema:name": req.dataset.publisher_name,
                "schema:url": {"@id": req.dataset.publisher_url}
            },
            "dc:license": {"schema:name": req.dataset.license}
        }
        if req.dataset.jsondata.get('license_url'):
            doc["dc:license"]["@id"] = req.dataset.jsondata.get('license_url')
        for k, v in kw.items():
            doc[k] = v
        return doc

    @classmethod
    def csvm_doc(cls, url, req, cols):
        primary_key = None
        foreign_keys = []
        columns = []

        for name, col in cols:
            spec = {
                'name': name,
                'datatype': {'@id': 'http://www.w3.org/2001/XMLSchema#string'}}
            if isinstance(col, property):
                col = None  # pragma: no cover
            else:
                if col is not None and not isinstance(col, Column):
                    try:
                        col = col.property.columns[0]
                    except AttributeError:  # pragma: no cover
                        col = None
                        raise
            if col is not None:
                if len(col.foreign_keys) == 1:
                    fk = list(col.foreign_keys)[0]
                    foreign_keys.append({
                        'columnReference': name,
                        'reference': {
                            'resource': fk.column.table.name,
                            'columnReference': fk.column.name}
                    })
                if col.primary_key:
                    primary_key = name
                for t, s in cls.type_map:
                    if isinstance(col.type, t):
                        spec['datatype']['@id'] = s
                        break
                if col.doc:
                    spec['dc:description'] = col.doc
            columns.append(spec)
        tableSchema = {'columns': columns}
        if primary_key:
            tableSchema['primaryKey'] = primary_key
        if foreign_keys:
            tableSchema['foreignKeys'] = foreign_keys
        return cls.csvm_basic_doc(req, url=url, tableSchema=tableSchema)

    def render(self, ctx, req):
        item = ctx.get_query(limit=1).first()
        fields = item.csv_head() if item else []
        cls = inspect(item).class_ if item else None
        doc = self.csvm_doc(
            req.url.replace('.csv-metadata.json', '.csv'),
            req,
            [(field, getattr(cls, field, None)) for field in fields])
        doc["dc:title"] = "{0} - {1}".format(req.dataset.name, ctx)
        return pyramid_render('json', doc, request=req)



"""Functionality to configure leaflet maps from python."""
from __future__ import unicode_literals, division, print_function, absolute_import
from six import string_types
from clldutils.misc import lazyproperty

from clld.interfaces import IDataTable, IMapMarker, IIcon
from clld.web.util import helpers
from clld.web.util.htmllib import HTML
from clld.web.util.component import Component
from clld.web.adapters.geojson import GeoJson, GeoJsonCombinationDomainElement, get_lonlat


class Layer(object):

    """Represents a layer in a leaflet map.

    A layer in our terminology is a
    `FeatureCollection <http://geojson.org/geojson-spec.html#feature-collection-objects>`_
    in geojson and a
    `geoJson layer <http://leafletjs.com/reference.html#geojson>`_
    in leaflet, i.e. a bunch of points on the map.
    """

    def __init__(self, id_, name, data, **kw):
        """Initialize a layer object.

        :param id_: Map-wide unique string identifying the layer.
        :param name: Human readable name of the layer.
        :param data: A GeoJSON FeatureCollection either specified as corresponding Python\
        dict or as URL which will serve the appropriate GeoJSON.
        :param kw: Additional keyword parameters are made available to the Layer as \
        instance attributes.
        """
        self.id = id_
        self.name = name
        self.data = data
        for k, v in kw.items():
            setattr(self, k, v)


class Legend(object):

    """Represents a navpill with a dropdown above a map."""

    def __init__(self,
                 map_,
                 name,
                 items,
                 label=None,
                 stay_open=False,
                 item_attrs=None,
                 pull_right=False):
        self.map = map_
        self.name = name
        self.label = label or name.capitalize()
        self.items = items
        self.stay_open = stay_open
        self.item_attrs = item_attrs or {}
        self.pull_right = pull_right

    def format_id(self, suffix=None):
        suffix = suffix or ''
        if suffix:
            suffix = '-' + suffix
        return 'legend-%s%s' % (self.name, suffix)

    def render_item(self, item):
        if not isinstance(item, (tuple, list)):
            item = [item]
        attrs = self.item_attrs
        if self.stay_open:
            class_ = attrs.get('class', attrs.get('class_', ''))
            attrs['class'] = class_ + ' stay-open'
        return HTML.li(*item, **attrs)

    def render(self):
        a_attrs = {
            'class': 'dropdown-toggle',
            'data-toggle': "dropdown",
            'href': "#",
            'id': self.format_id('opener')}
        ul_class = 'dropdown-menu'
        if self.stay_open:
            ul_class += ' stay-open'
        return HTML.li(
            HTML.a(self.label, HTML.b(class_='caret'), **a_attrs),
            HTML.ul(
                *map(self.render_item, self.items),
                **dict(class_=ul_class, id=self.format_id('container'))),
            class_='dropdown' + (' pull-right' if self.pull_right else ''),
            id=self.format_id(),
        )


class FilterLegend(Legend):

    """Legend with actionable items.

    Legend rendering radio controls to filter languages on a map in sync with a column
    of an associated DataTable.
    """

    def __init__(self, map_, value_getter, col=None, dt=None, **kw):
        """Initialize.

        @param value_getter: Name of a javascript object which will be called with the \
        properties associated with a map marker to determine its filter value.
        """
        kw.setdefault('stay_open', True)
        if col and dt:
            col = {c.name: c for c in dt.cols}.get(col)
        if col:
            kw.setdefault('label', col.js_args.get('sTitle'))
        Legend.__init__(self, map_, col.name if col else 'nocol', [], **kw)
        self.jsname = 'fl-' + self.name
        items = [self.li(col, '--any--', value_getter, checked=True)]
        for item in getattr(col, 'choices', []):
            items.append(self.li(col, item, value_getter))
        self.items = items

    def li_label(self, item):
        return item[1] if isinstance(item, (tuple, list)) else item

    def li(self, col, item, value_getter, checked=False):
        input_attrs = dict(
            type='radio',
            class_='stay-open %s inline' % self.jsname,
            name=self.jsname,
            value=item[0] if isinstance(item, (tuple, list)) else item,
            onclick=helpers.JS("CLLD.mapLegendFilter")(
                self.map.eid,
                self.name,
                self.jsname,
                helpers.JS(value_getter),
                col.dt.eid if col else None))
        if checked:
            input_attrs['checked'] = 'checked'
        return HTML.label(
            HTML.input(**input_attrs),
            ' ',
            self.li_label(item),
            class_="stay-open",
            style="margin-left:5px; margin-right:5px;",
        )


class Map(Component):

    """Represents the configuration for a leaflet map."""

    __template__ = 'clld:web/templates/map.mako'

    def __init__(self, ctx, req, eid='map'):
        """Initialize.

        :param ctx: context object of the current request.
        :param req: current pyramid request object.
        :param eid: Page-unique DOM-node ID.
        """
        self.req = req
        self.ctx = ctx
        self.eid = eid
        self.map_marker = req.registry.getUtility(IMapMarker)

    def get_options_from_req(self):
        params = self.req.params
        res = {}
        try:
            if 'lat' in params and 'lng' in params:
                res['center'] = list(map(float, [params['lat'], params['lng']]))
            if 'z' in params:
                res['zoom'] = int(params['z'])
        except (ValueError, TypeError):
            pass
        return res

    @lazyproperty
    def layers(self):
        """The list of layers of the map.

        .. note:: Since layers may be costly to compute, we cache them per map instance.

        :return: list of :py:class:`clld.web.maps.Layer` instances.
        """
        return list(self.get_layers())

    def get_layers(self):
        """Generate the list of layers.

        :return: list or generator of :py:class:`clld.web.maps.Layer` instances.
        """
        route_params = {'ext': 'geojson'}
        if not IDataTable.providedBy(self.ctx):
            route_params['id'] = self.ctx.id
        route_name = self.req.matched_route.name
        if not route_name.endswith('_alt'):
            route_name += '_alt'
        yield Layer(
            getattr(self.ctx, 'id', 'id'),
            '%s' % self.ctx,
            self.req.route_url(route_name, **route_params))

    @lazyproperty
    def legends(self):
        return list(self.get_legends())

    def get_legends(self):
        if len(self.layers) > 1:
            items = []
            total = 0
            repr_attrs = dict(class_='pull-right stay-open', style="padding-right: 10px;")

            for layer in self.layers:
                representation = ''
                if hasattr(layer, 'representation'):
                    total += layer.representation
                    representation = HTML.span(str(layer.representation), **repr_attrs)
                items.append([
                    HTML.label(
                        HTML.input(
                            class_="stay-open",
                            type="checkbox",
                            checked="checked",
                            onclick=helpers.JS_CLLD.mapToggleLayer(
                                self.eid, layer.id, helpers.JS("this"))),
                        getattr(layer, 'marker', ''),
                        layer.name,
                        class_="checkbox inline stay-open",
                        style="margin-left: 5px; margin-right: 5px;",
                    ),
                    representation,
                ])
            if total:
                items.append(HTML.span(HTML.b(str(total)), **repr_attrs))
            yield Legend(
                self,
                'layers',
                items,
                label='Legend',
                stay_open=True,
                item_attrs=dict(style='clear: right'))
        items = []
        for size in [15, 20, 30, 40]:
            attrs = dict(name="iconsize", value=str(size), type="radio")
            if size == self.options.get('icon_size', 30):
                attrs['checked'] = 'checked'
            items.append(HTML.label(
                HTML.input(onclick=helpers.JS_CLLD.mapResizeIcons(self.eid), **attrs),
                HTML.img(
                    height=str(size),
                    width=str(size),
                    src=self.req.registry.getUtility(IIcon, 'cff6600').url(self.req)),
                class_="radio",
                style="margin-left: 5px; margin-right: 5px;"))
        yield Legend(
            self,
            'iconsize',
            items,
            label='Icon size')

        def item(layer):
            return HTML.a(
                layer.name,
                onclick='return %s;' % helpers.JS_CLLD.mapShowGeojson(self.eid, layer.id),
                href=layer.data if isinstance(layer.data, string_types) else '#')
        yield Legend(
            self, 'geojson', map(item, self.layers), label='GeoJSON', pull_right=True)


class ParameterMap(Map):

    """Map displaying markers for valuesets associated with a parameter instance."""

    def get_layers(self):
        if self.ctx.domain:
            for de in self.ctx.domain:
                yield Layer(
                    de.id,
                    de.name,
                    self.req.resource_url(
                        self.ctx, ext='geojson',
                        _query=dict(domainelement=str(de.id), **self.req.query_params)
                    ),
                    marker=helpers.map_marker_img(self.req, de, marker=self.map_marker))
        else:
            yield Layer(
                self.ctx.id,
                self.ctx.name,
                self.req.resource_url(self.ctx, ext='geojson'))

    def get_default_options(self):
        return {'info_query': {'parameter': self.ctx.pk}, 'hash': True}


class GeoJsonMultiple(GeoJson):

    """Render a collection of languages as geojson feature collection."""

    def feature_iterator(self, ctx, req):
        return ctx

    def feature_properties(self, ctx, req, language):
        return {
            'icon': req.registry.getUtility(IIcon, 'tff0000').url(req),
            'icon_size': 10,
            'zindex': 1000}


class CombinationMap(Map):

    """Map for a combination of parameters."""

    def get_layers(self):
        for de in self.ctx.domain:
            if de.languages:
                yield Layer(
                    de.id,
                    de.name,
                    GeoJsonCombinationDomainElement(de).render(de, self.req, dump=False),
                    marker=HTML.img(src=de.icon.url(self.req), height='20', width='20'))
        if self.ctx.multiple:
            # yield another layer which can be used to mark languages with multiple
            # values, because this may not be visible when markers are stacked on top
            # of each other.
            icon_url = self.req.registry.getUtility(IIcon, 'tff0000').url(self.req)
            yield Layer(
                '__multiple__',
                'Languages with multiple values',
                GeoJsonMultiple(None).render(self.ctx.multiple, self.req, dump=False),
                marker=HTML.img(src=icon_url, height='20', width='20'))

    def get_options(self):
        return {'icon_size': 25, 'hash': True}


class LanguageMap(Map):

    """Map showing a single language."""

    def get_layers(self):
        yield Layer(
            self.ctx.id,
            self.ctx.name,
            GeoJson(self.ctx).render(self.ctx, self.req, dump=False))

    def get_default_options(self):
        return {
            'center': list(reversed(get_lonlat(self.ctx) or [0, 0])),
            'zoom': 3,
            'no_popup': True,
            'no_link': True,
            'sidebar': True}


class GeoJsonSelectedLanguages(GeoJson):

    """Represents the geo-data of an iterable selection of languages.

    The iterable is assumed to be passed into the adapter upon initialization.
    """

    def feature_iterator(self, ctx, req):
        return self.obj


class SelectedLanguagesMap(Map):

    """Map showing an arbitrary selection of languages."""

    def __init__(self, ctx, req, languages, geojson_impl=None, **kw):
        """Initialize.

        :param languages: Iterable collection of languages.
        :param geojson_impl: GeoJson implementation to use.
        """
        self.geojson_impl = geojson_impl or GeoJsonSelectedLanguages
        self.languages = languages
        Map.__init__(self, ctx, req, **kw)

    def get_options(self):
        return {'icon_size': 20, 'hash': True, 'show_labels': len(self.languages) < 100}

    def get_layers(self):
        yield Layer(
            'languages',
            'Languages',
            self.geojson_impl(self.languages).render(self.ctx, self.req, dump=False))


#
# The following code implements a map to overlay geojson for parameters from distinct
# datasets. It may be used by CrossGram at some point.
#
def layers(spec, size, zindex=0):  # pragma: no cover
    import requests

    app, pid, url = spec

    def normalize(geojson):
        for f in geojson['features']:
            f['properties']['icon_size'] = size
            f['properties']['zindex'] = zindex
            if f['geometry']['coordinates'][0] > 180:
                f['geometry']['coordinates'][0] = f['geometry']['coordinates'][0] - 360
        return geojson

    geojson = requests.get(url + '.geojson').json()
    id_ = '-'.join([app, pid])

    if geojson['properties']['domain'] and not geojson['features']:
        for de in geojson['properties']['domain']:
            yield Layer(
                '-'.join([id_, de['id']]),
                '%s: %s - %s' % (app, geojson['properties']['name'], de['name']),
                normalize(
                    requests.get(url + '.geojson?domainelement=' + de['id']).json()),
                size=size,
                link=url,
                marker=HTML.img(src=de['icon'], width=size, height=size))
    else:
        yield Layer(
            id_,
            '%s: %s' % (app, geojson['properties']['name']),
            normalize(geojson),
            size=size,
            link=url,
            domain=geojson['properties']['domain'])


class CombinedMap(Map):  # pragma: no cover

    """Map for combination of parameters from different clld apps."""

    def get_layers(self):
        for i, spec in enumerate(self.ctx):
            for layer in layers(spec, (i + 1) * 10 + 10, (i + 1) * (-1000)):
                yield layer

    def get_options(self):
        return {'no_popup': True, 'no_link': True}
